import os
import json
from pathlib import Path


function_groups = {
    ('src', 'include/duckdb', 'function'): ['scalar', 'aggregate'],
    ('extension', 'core_functions/include', 'core_functions'): ['scalar', 'aggregate'],
}


def get_header():
    return '''//===----------------------------------------------------------------------===//
//                         DuckDB
//
// {HEADER}_functions.hpp
//
//
//===----------------------------------------------------------------------===//
// This file is automatically generated by scripts/generate_functions.py
// Do not edit this file manually, your changes will be overwritten
//===----------------------------------------------------------------------===//

#pragma once

#include "duckdb/function/function_set.hpp"

namespace duckdb {

'''


def get_footer():
    return '''} // namespace duckdb
'''


def main():
    function_type_set = {}
    for (root, include_dir, group), function_types in sorted(function_groups.items()):
        all_functions_group = []
        group_dir = Path(group)
        for function_type in function_types:
            type_dir = Path(root).joinpath(group_dir.joinpath(function_type))
            relative_function_paths = sorted(
                [f'{group}/{function_type}/{f.name}' for f in type_dir.iterdir() if f.is_dir()]
            )
            for function_path in relative_function_paths:
                if Path(normalize_path_separators(f'{root}/{function_path}/functions.json')).exists():
                    create_header_file(root, include_dir, function_path, all_functions_group, function_type_set)
        create_function_list_file(root, group, all_functions_group)


def normalize_path_separators(x):
    return os.path.sep.join(x.split('/'))


def legal_struct_name(name):
    return name.isalnum()


def get_struct_name(function_name):
    return function_name.replace('_', ' ').title().replace(' ', '') + 'Fun'


def get_parameter_line(variants):
    if not all(
        isinstance(variant['parameters'], list)
        and all(isinstance(param, dict) for param in variant['parameters'])
        and all('name' in param.keys() for param in variant['parameters'])
        for variant in variants
    ):
        raise ValueError(
            f"invalid parameters for variants {variants}\nParameters should have format: \"parameters\": [{{\"name\": <param_name>, \"type\": <param_type>}}, ...]"
        )
    return "\\001".join(
        ",".join(
            param['name'] + "::" + param['type'] if ('type' in param) else param['name']
            for param in variant['parameters']
        )
        for variant in variants
    )


def get_description_line(variants):
    return "\\001".join([variant['description'] for variant in variants])


def get_example_line(variants):
    return "\\001".join([example_from_json(variant) for variant in variants])


def example_from_json(json_record):
    if 'example' in json_record:
        example_line = sanitize_string(json_record['example'])
    elif 'examples' in json_record:
        example_line = examples_to_line(json_record['examples'])
    else:
        example_line = ''
    return example_line


def examples_to_line(example_list):
    return "\\002".join([sanitize_string(example) for example in example_list])


def get_category_line(variants):
    return "\\001".join([categories_from_json(variant) for variant in variants])


def categories_from_json(json_record):
    if 'categories' in json_record:
        category_line = ','.join([category.strip() for category in json_record['categories']])
    else:
        category_line = ''
    return category_line


def sanitize_string(text):
    return text.replace('\\', '\\\\').replace('"', '\\"')


def create_header_file(root, include_dir, path, all_function_list, function_type_set):
    header_path = normalize_path_separators(f'{root}/{include_dir}/{path}_functions.hpp')
    json_path = normalize_path_separators(f'{root}/{path}/functions.json')
    with open(json_path, 'r') as f:
        parsed_json = json.load(f)
    new_text = get_header().replace('{HEADER}', path)
    for entry in parsed_json:
        function_text = ''
        if 'struct' in entry:
            struct_name = entry['struct']
        else:
            struct_name = get_struct_name(entry['name'])
        if not legal_struct_name(struct_name):
            print(f'Struct name {struct_name} is not a valid struct name!')
            exit(1)
        if struct_name in function_type_set:
            raise Exception("Duplicate entry " + struct_name)
        function_type_set[struct_name] = entry['type']
        if entry['type'] == 'scalar_function':
            function_text = 'static ScalarFunction GetFunction();'
            all_function_list.append([entry['name'], f"DUCKDB_SCALAR_FUNCTION({struct_name})"])
        elif entry['type'] == 'scalar_function_set':
            function_text = 'static ScalarFunctionSet GetFunctions();'
            all_function_list.append([entry['name'], f"DUCKDB_SCALAR_FUNCTION_SET({struct_name})"])
        elif entry['type'] == 'aggregate_function':
            function_text = 'static AggregateFunction GetFunction();'
            all_function_list.append([entry['name'], f"DUCKDB_AGGREGATE_FUNCTION({struct_name})"])
        elif entry['type'] == 'aggregate_function_set':
            function_text = 'static AggregateFunctionSet GetFunctions();'
            all_function_list.append([entry['name'], f"DUCKDB_AGGREGATE_FUNCTION_SET({struct_name})"])
        else:
            print("Unknown entry type " + entry['type'] + ' for entry ' + struct_name)
            exit(1)
        if 'variants' in entry:
            parameter_line = get_parameter_line(entry['variants'])
            description_line = get_description_line(entry['variants'])
            example_line = get_example_line(entry['variants'])
            category_line = get_category_line(entry['variants'])
        else:
            parameter_line = entry['parameters'].replace(' ', '') if 'parameters' in entry else ''
            description_line = sanitize_string(entry['description'])
            example_line = example_from_json(entry)
            category_line = categories_from_json(entry)
        if 'extra_functions' in entry:
            for func_text in entry['extra_functions']:
                function_text += '\n	' + func_text
        new_text += (
            '''struct {STRUCT} {
	static constexpr const char *Name = "{NAME}";
	static constexpr const char *Parameters = "{PARAMETERS}";
	static constexpr const char *Description = "{DESCRIPTION}";
	static constexpr const char *Example = "{EXAMPLE}";
	static constexpr const char *Categories = "{CATEGORIES}";

	{FUNCTION}
};

'''.replace(
                '{STRUCT}', struct_name
            )
            .replace('{NAME}', entry['name'])
            .replace('{PARAMETERS}', parameter_line)
            .replace('{DESCRIPTION}', description_line)
            .replace('{EXAMPLE}', example_line)
            .replace('{CATEGORIES}', category_line)
            .replace('{FUNCTION}', function_text)
        )
        alias_count = 1
        if 'aliases' in entry:
            for alias in entry['aliases']:
                alias_struct_name = get_struct_name(alias)
                if not legal_struct_name(alias_struct_name):
                    alias_struct_name = struct_name + 'Alias'
                    if alias_count > 1:
                        alias_struct_name += str(alias_count)
                    alias_count += 1

                aliased_type = entry['type']
                if aliased_type == 'scalar_function':
                    all_function_list.append([alias, f"DUCKDB_SCALAR_FUNCTION_ALIAS({alias_struct_name})"])
                elif aliased_type == 'scalar_function_set':
                    all_function_list.append([alias, f"DUCKDB_SCALAR_FUNCTION_SET_ALIAS({alias_struct_name})"])
                elif aliased_type == 'aggregate_function':
                    all_function_list.append([alias, f"DUCKDB_AGGREGATE_FUNCTION_ALIAS({alias_struct_name})"])
                elif aliased_type == 'aggregate_function_set':
                    all_function_list.append([alias, f"DUCKDB_AGGREGATE_FUNCTION_SET_ALIAS({alias_struct_name})"])
                else:
                    print("Unknown entry type " + aliased_type + ' for entry ' + struct_name)
                    exit(1)
                function_type_set[alias_struct_name] = aliased_type
                new_text += (
                    '''struct {STRUCT} {
	using ALIAS = {ALIAS};

	static constexpr const char *Name = "{NAME}";
};

'''.replace(
                        '{STRUCT}', alias_struct_name
                    )
                    .replace('{NAME}', alias)
                    .replace('{ALIAS}', struct_name)
                )
    new_text += get_footer()
    with open(header_path, 'w+') as f:
        f.write(new_text)


def create_function_list_file(root, group, all_function_list):
    function_list_file = normalize_path_separators(f'{root}/{group}/function_list.cpp')
    with open(function_list_file, 'r') as f:
        text = f.read()

    static_function = f'static const StaticFunctionDefinition {group}[]' ' = {'
    pos = text.find(static_function)
    header = text[:pos]
    footer_lines = text[pos:].split('\n')
    footer = ''
    for i in range(len(footer_lines)):
        if len(footer_lines[i]) == 0:
            footer = '\n'.join(footer_lines[i:])
            break

    new_text = header
    new_text += static_function + '\n'
    all_function_list = sorted(all_function_list, key=lambda x: x[0])
    for entry in all_function_list:
        new_text += '\t' + entry[1] + ',\n'
    new_text += '\tFINAL_FUNCTION\n};\n'
    new_text += footer

    with open(function_list_file, 'w+') as f:
        f.write(new_text)


if __name__ == "__main__":
    main()



# fmt: off

import pytest
import subprocess
import sys
from typing import List
from conftest import ShellTest
import os

@pytest.mark.skip(reason="Skip after File Logging rework")
def test_http_logging_file(shell, tmp_path):
    temp_dir = tmp_path / 'http_logging_dir'
    temp_dir.mkdir()
    temp_file = temp_dir / 'myfile'

    test = (
        ShellTest(shell)
        .statement("SET enable_http_logging=true;")
        .statement(f"SET http_logging_output='{temp_file.as_posix()}'")
        .statement("install 'http://extensions.duckdb.org/v0.10.1/osx_arm64/httpfs.duckdb_extension.gzzz';")
    )
    result = test.run()

    with open(temp_file, 'r') as f:
        file_content = f.read()
        assert "HTTP Request" in file_content
        assert "HTTP Response" in file_content



# fmt: off

import pytest
from conftest import ShellTest

def test_pager_status_default(shell):
    """Test that pager status shows 'automatic' by default"""
    test = (
        ShellTest(shell)
            .statement('.pager')
    )
    result = test.run()
    result.check_stdout('Pager mode: automatic')


def test_pager_help(shell):
    """Test that .help shows pager documentation"""
    test = (
        ShellTest(shell)
            .statement('.help pager')
    )
    result = test.run()
    result.check_stdout('DUCKDB_PAGER')


def test_pager_off_explicit(shell):
    """Test setting pager explicitly to off"""
    test = (
        ShellTest(shell)
            .statement('.pager off')
            .statement('.pager')
    )
    result = test.run()
    result.check_stdout('Pager mode: off')

def test_pager_on_with_pager_env(shell):
    """Test that pager uses PAGER environment variable"""
    test = (
        ShellTest(shell)
            .statement('.pager on')
            .statement('.pager')
    )
    test.environment['PAGER'] = 'less'
    result = test.run()
    result.check_stdout('less')


def test_pager_on_with_duckdb_pager_env(shell):
    """Test that DUCKDB_PAGER takes precedence over PAGER"""
    test = (
        ShellTest(shell)
            .statement('.pager on')
            .statement('.pager')
    )
    test.environment['PAGER'] = 'less'
    test.environment['DUCKDB_PAGER'] = 'less -SR'
    result = test.run()
    result.check_stdout('less -SR')


def test_pager_duckdb_pager_priority(shell):
    """Test DUCKDB_PAGER shows in status even when pager is off"""
    test = (
        ShellTest(shell)
            .statement('.pager on')
            .statement('.pager')
    )
    test.environment['DUCKDB_PAGER'] = 'less -R'
    result = test.run()
    result.check_stdout('less -R')


def test_pager_custom_command(shell):
    """Test setting a custom pager command"""
    test = (
        ShellTest(shell)
            .statement(".pager 'cat'")
            .statement('.pager')
    )
    result = test.run()
    result.check_stdout('cat')


def test_pager_custom_command_with_args(shell):
    """Test setting a custom pager command with arguments"""
    test = (
        ShellTest(shell)
            .statement(".pager 'less -SR'")
            .statement('.pager')
    )
    result = test.run()
    result.check_stdout('less -SR')

def test_pager_with_query_output(shell):
    """Test that pager works with query output using cat"""
    test = (
        ShellTest(shell)
            .statement(".mode csv")
            .statement(".pager 'cat'")
            .statement('FROM range(10000)')
    )
    result = test.run()
    result.check_stdout('8888')

def test_pager_doesnt_affect_error_messages(shell):
    """Test that pager doesn't capture error messages"""
    test = (
        ShellTest(shell)
            .statement(".pager 'cat'")
            .statement(".mode csv")
            .statement('SELECT invalid_column FROM nonexistent_table')
    )
    result = test.run()
    result.check_stderr('Table')


def test_pager_preserves_nullvalue(shell):
    """Test that pager preserves null value rendering"""
    test = (
        ShellTest(shell)
            .statement('.nullvalue XYZ')
            .statement(".pager 'cat'")
            .statement('SELECT NULL FROM range(10000)')
    )
    result = test.run()
    result.check_stdout('XYZ')

def test_pager_multiple_queries(shell):
    """Test pager with multiple queries in sequence"""
    test = (
        ShellTest(shell)
            .statement(".pager 'cat'")
            .statement(".mode csv")
            .statement('SELECT 1 FROM range(1000)')
            .statement('SELECT 2 FROM range(1000)')
            .statement('SELECT 3 FROM range(1000)')
    )
    result = test.run()
    result.check_stdout('1')
    result.check_stdout('2')
    result.check_stdout('3')

def test_pager_small_data(shell):
    """Test pager with a small data set"""
    test = (
        ShellTest(shell)
            .statement(".pager 'unknown_cmd'")
            .statement(".mode csv")
            .statement('FROM range(10)')
    )
    result = test.run()
    result.check_stdout('9')

# fmt: on



import json
import os
import sys
import glob
import mimetypes
import urllib.request

api_url = 'https://api.github.com/repos/duckdb/duckdb/'

if len(sys.argv) < 2:
    print("Usage: [filename1] [filename2] ... ")
    exit(1)

# this essentially should run on release tag builds to fill up release assets and main

pr = os.getenv("TRAVIS_PULL_REQUEST", "")
if pr != "false":
    print("Not running on PRs. Exiting.")
    exit(0)

tag = os.getenv("TRAVIS_TAG", '')  # this env var is always present just not always used
if tag == '':
    tag = 'main-builds'
print("Running on tag %s" % tag)

if tag == "main-builds" and os.getenv("TRAVIS_BRANCH", "") != "main":
    print("Only running on main branch for %s tag. Exiting." % tag)
    exit(0)


token = os.getenv("GH_TOKEN", "")
if token == "":
    raise ValueError('need a GitHub token in GH_TOKEN')


def gh_api(suburl, filename='', method='GET'):
    url = api_url + suburl
    headers = {"Content-Type": "application/json", 'Authorization': 'token ' + token}

    body_data = b''

    if len(filename) > 0:
        method = 'POST'
        body_data = open(filename, 'rb')

        mime_type = mimetypes.guess_type(local_filename)[0]
        if mime_type is None:
            mime_type = "application/octet-stream"
        headers["Content-Type"] = mime_type
        headers["Content-Length"] = os.path.getsize(local_filename)

        url = suburl  # cough

    req = urllib.request.Request(url, body_data, headers)
    req.get_method = lambda: method
    try:
        raw_resp = urllib.request.urlopen(req).read().decode()
    except urllib.error.HTTPError as e:
        raw_resp = e.read().decode()  # gah

    if method != 'DELETE':
        return json.loads(raw_resp)
    else:
        return {}


# check if tag exists
resp = gh_api('git/ref/tags/%s' % tag)
if 'object' not in resp or 'sha' not in resp['object']:  # or resp['object']['sha'] != sha
    raise ValueError('tag %s not found' % tag)

resp = gh_api('releases/tags/%s' % tag)
if 'id' not in resp or 'upload_url' not in resp:
    raise ValueError('release does not exist for tag ' % tag)

# double-check that release exists and has correct sha
# disabled to not spam people watching releases
# if 'id' not in resp or 'upload_url' not in resp or 'target_commitish' not in resp or resp['target_commitish'] != sha:
# 	raise ValueError('release does not point to requested commit %s' % sha)

# TODO this could be a paged response!
assets = gh_api('releases/%s/assets' % resp['id'])

upload_url = resp['upload_url'].split('{')[0]  # gah
files = sys.argv[1:]
for filename in files:
    if '=' in filename:
        parts = filename.split("=")
        asset_filename = parts[0]
        paths = glob.glob(parts[1])
        if len(paths) != 1:
            raise ValueError("Could not find file for pattern %s" % local_filename)
        local_filename = paths[0]
    else:
        asset_filename = os.path.basename(filename)
        local_filename = filename

    # delete if present
    for asset in assets:
        if asset['name'] == asset_filename:
            gh_api('releases/assets/%s' % asset['id'], method='DELETE')

    resp = gh_api(upload_url + '?name=%s' % asset_filename, filename=local_filename)
    if 'id' not in resp:
        raise ValueError('upload failed :/ ' + str(resp))
    print("%s -> %s" % (local_filename, resp['browser_download_url']))



# this script re-generates the storage used for storage_version.test_slow
# before running this script, increment the version number in src/storage/storage_info.cpp and recompile (`make`)

import os
import subprocess
from python_helpers import open_utf8

shell_proc = os.path.join('build', 'release', 'duckdb')

gen_storage_script = os.path.join('test', 'sql', 'storage_version', 'generate_storage_version.sql')
gen_storage_target = os.path.join('test', 'sql', 'storage_version', 'storage_version.db')


def try_remove_file(fname):
    try:
        os.remove(fname)
    except:
        pass


try_remove_file(gen_storage_target)
try_remove_file(gen_storage_target + '.wal')


def run_command_in_shell(cmd):
    print(cmd)
    res = subprocess.run(
        [shell_proc, '--batch', '-no-init', gen_storage_target],
        capture_output=True,
        input=bytearray(cmd, 'utf8'),
    )
    stdout = res.stdout.decode('utf8').strip()
    stderr = res.stderr.decode('utf8').strip()
    if res.returncode != 0:
        print("Failed to create database file!")
        print("----STDOUT----")
        print(stdout)
        print("----STDERR----")
        print(stderr)


with open_utf8(gen_storage_script, 'r') as f:
    cmd = f.read()

run_command_in_shell(cmd)
run_command_in_shell('select * from integral_values')
run_command_in_shell('select * from integral_values')

try_remove_file(gen_storage_target + '.wal')



from datetime import datetime
import time
from utils.logger import make_logger
import os
import json
from difflib import unified_diff

logger = make_logger(__name__)


class TestReport:
    step_names = ["serialize_queries_plans", "execute_all_plans_from_file", "serialize_results", "compare_results"]

    def __init__(self, test_spec, old_version, new_version, exit_on_failure=False):
        self.test_filename = test_spec.test_absolute_filename
        self.test_relative_path = test_spec.test_spec_relative_path
        self.test_runtime_directory = test_spec.test_runtime_directory  # Test execution directory
        self.comparison_results_file_name = f"{test_spec.results_new_file_name}.report"
        self.comparison_results = None
        self.start_time = time.time()
        self.exit_on_failure = exit_on_failure
        self.last_timestamp = self.start_time
        self.durations = []
        self.outputs = []
        self.duckdb_old_version = old_version
        self.duckdb_new_version = new_version

    def end_cached_step(self, step_name, *args):
        str_args = ", ".join([f"'{arg}'" for arg in args])
        query = f"CALL {step_name}({str_args});"
        return self.end_step(step_name, {"success": True, "output": ["cached"], "error": None, "query": query})

    def end_step(self, step_name, step_output):
        if self.last_timestamp is None:
            raise ValueError("Test has not been started.")

        if step_name not in TestReport.step_names:
            raise ValueError(f"Unknown step name: {step_name}")

        self.durations.append(time.time() - self.last_timestamp)
        self.outputs.append(step_output)
        self.last_timestamp = time.time()
        success = step_output["success"]

        if self.exit_on_failure and not success:
            logger.error(f"Exiting following failure at step '{step_name}' failed for test '{self.test_filename}'")

            queries = '\n'.join([output['query'] for output in self.outputs])
            logger.info(f"Previous queries:\n{queries}")
            exit()

        return success

    def log_errors(self):
        for output in self.outputs:
            if 'error_log' in output:
                logger.error(output['error_log'])

    def log_steps_queries(self):
        steps = [f".cd {self.test_runtime_directory}\nLOAD 'test_utils';"]
        for output in self.outputs:
            if 'query' not in output:
                print(f"Output does not contain 'query': {output}")
            steps.append(output['query'])

        logger.info("Executed steps:\n" + "\n".join(steps))

    def find_exception_message(self):
        for output in self.outputs:
            if 'exception_message' in output:
                return output['exception_message']
        return '** No exception message found **'

    def load_comparison_results(self):
        # Check if `self.comparison_results_file_name` exists
        if not os.path.exists(self.comparison_results_file_name):
            return

        # Load as JSON
        with open(self.comparison_results_file_name, 'r') as f:
            self.comparison_results = json.load(f)

            error_count_map = {}
            for comparison_error in self.comparison_results:
                err_msg = comparison_error['error']
                if err_msg not in error_count_map:
                    error_count_map[err_msg] = 0
                error_count_map[err_msg] += 1

                expected = (
                    comparison_error['expected'].splitlines(keepends=True) if 'expected' in comparison_error else []
                )
                actual = comparison_error['actual'].splitlines(keepends=True) if 'actual' in comparison_error else []
                comparison_error['diff'] = ''.join(
                    unified_diff(expected, actual, fromfile='expected', tofile='actual', n=0)
                )

            for err_msg, count in error_count_map.items():
                if count == 1:
                    logger.warning(err_msg)
                else:
                    logger.warning(f"{err_msg} ({count})")

    def is_successful(self):
        return all(output["success"] for output in self.outputs)

    @staticmethod
    def report_schema():
        schema = """
    CREATE TABLE test_report (
      test_filename VARCHAR,
      test_runtime_directory VARCHAR,
      start_time TIMESTAMP,
      duckdb_old_version VARCHAR,
      duckdb_new_version VARCHAR,
    """
        for step_name in TestReport.step_names:
            schema += f"  {step_name}_duration DOUBLE,\n"
            schema += f"  {step_name}_outcome VARCHAR,\n"
            schema += f"  {step_name}_stdout VARCHAR,\n"
            schema += f"  {step_name}_stderr VARCHAR,\n"

        schema += f"  comparison_report VARCHAR,\n"
        schema += ");\n"
        return schema

    @staticmethod
    def report_insert():
        sql = f"INSERT INTO test_report (test_filename, test_runtime_directory, start_time, duckdb_old_version, duckdb_new_version, "
        sql += ", ".join(
            [
                f"{step_name}_duration, {step_name}_outcome, {step_name}_stdout, {step_name}_stderr"
                for step_name in TestReport.step_names
            ]
        )
        sql += ", comparison_report"
        sql += ") VALUES (?, ?, ?, ?, ?, " + ("?, ?, ?, ?, " * len(TestReport.step_names)) + "?);\n"
        return sql

    def report_sql_values(self):
        values = [
            self.test_filename,
            self.test_runtime_directory,
            datetime.fromtimestamp(self.start_time),
            self.duckdb_old_version,
            self.duckdb_new_version,
        ]
        for i in range(len(TestReport.step_names)):
            if i < len(self.durations):
                values.append(self.durations[i])

                output = self.outputs[i]
                values.append("success" if output["success"] else "failure")

                # Remove 'result' and 'true' if they are the last two lines
                if output["output"] and len(output["output"]) > 1 and output["output"][-2:] == ["result", "true"]:
                    output["output"] = output["output"][:-2]

                values.append(", ".join(output["output"]))
                values.append(output["error"])
            else:
                values.append(None)
                values.append("not_run")
                values.append("")
                values.append("")

        if self.comparison_results is not None:
            values.append(json.dumps(self.comparison_results, ensure_ascii=False))
        else:
            values.append(None)
        return values



# this script generates data for the TPC-H dbgen
import os
from python_helpers import open_utf8


def get_csv_text(fpath, add_null_terminator=False):
    with open(fpath, 'rb') as f:
        text = bytearray(f.read())
    result_text = ""
    first = True
    for byte in text:
        if first:
            result_text += str(byte)
        else:
            result_text += ", " + str(byte)
        first = False
    if add_null_terminator:
        result_text += ", 0"
    return result_text


def write_dir(dirname, varname):
    files = os.listdir(dirname)
    files.sort()
    result = ""
    aggregated_result = "const char *%s[] = {\n" % (varname,)
    for fname in files:
        file_varname = "%s_%s" % (varname, fname.split('.')[0])
        result += "const uint8_t %s[] = {" % (file_varname,) + get_csv_text(os.path.join(dirname, fname), True) + "};\n"
        aggregated_result += "\t(const char*) %s,\n" % (file_varname,)
    aggregated_result = aggregated_result[:-2] + "\n};\n"
    return result + aggregated_result


# ------------------------------------------- #
# ------------------------------------------- #
# -------------      TPC-H       ------------ #
# ------------------------------------------- #
# ------------------------------------------- #
tpch_dir = 'extension/tpch/dbgen'
tpch_queries = os.path.join(tpch_dir, 'queries')
tpch_answers_sf001 = os.path.join(tpch_dir, 'answers', 'sf0.01')
tpch_answers_sf01 = os.path.join(tpch_dir, 'answers', 'sf0.1')
tpch_answers_sf1 = os.path.join(tpch_dir, 'answers', 'sf1')
tpch_header = os.path.join(tpch_dir, 'include', 'tpch_constants.hpp')


def create_tpch_header(tpch_dir):
    result = """/* THIS FILE WAS AUTOMATICALLY GENERATED BY generate_csv_header.py */

#pragma once

const int TPCH_QUERIES_COUNT = 22;
"""
    # write the queries
    result += write_dir(tpch_queries, "TPCH_QUERIES")
    result += write_dir(tpch_answers_sf001, "TPCH_ANSWERS_SF0_01")
    result += write_dir(tpch_answers_sf01, "TPCH_ANSWERS_SF0_1")
    result += write_dir(tpch_answers_sf1, "TPCH_ANSWERS_SF1")

    with open_utf8(tpch_header, 'w+') as f:
        f.write(result)


print(tpch_header)
create_tpch_header(tpch_dir)

# ------------------------------------------- #
# ------------------------------------------- #
# -------------      TPC-DS      ------------ #
# ------------------------------------------- #
# ------------------------------------------- #
tpcds_dir = 'extension/tpcds/dsdgen'
tpcds_queries = os.path.join(tpcds_dir, 'queries')
tpcds_answers_sf001 = os.path.join(tpcds_dir, 'answers', 'sf0.01')
tpcds_answers_sf1 = os.path.join(tpcds_dir, 'answers', 'sf1')
tpcds_header = os.path.join(tpcds_dir, 'include', 'tpcds_constants.hpp')


def create_tpcds_header(tpch_dir):
    result = """/* THIS FILE WAS AUTOMATICALLY GENERATED BY generate_csv_header.py */

#pragma once

const int TPCDS_QUERIES_COUNT = 99;
const int TPCDS_TABLE_COUNT = 24;
"""
    # write the queries
    result += write_dir(tpcds_queries, "TPCDS_QUERIES")
    result += write_dir(tpcds_answers_sf001, "TPCDS_ANSWERS_SF0_01")
    result += write_dir(tpcds_answers_sf1, "TPCDS_ANSWERS_SF1")

    with open_utf8(tpcds_header, 'w+') as f:
        f.write(result)


print(tpcds_header)
create_tpcds_header(tpcds_dir)



from typing import Any, Generator, List
import json
import os
import time
from zipfile import Path
from os.path import basename, dirname
import shutil

import duckdb

from duckdb_sqllogictest import SQLLogicParser
from duckdb_sqllogictest.result import (
    ExecuteResult,
    Foreach,
    Load,
    Loop,
    Query,
    Require,
    RequireResult,
    Restart,
    Set,
    SQLLogicContext,
    SQLLogicRunner,
    Statement,
    Unzip,
)

from duckdb_sqllogictest.expected_result import ExpectedResult
from duckdb_sqllogictest.statement.query import SortStyle

from utils.query_exclusions import format_exclusion_tags
from utils.logger import make_logger

BWC_TAG_PREFIX = '-- bwc_tag:'
BWC_OUTPUT_DIR_NAME = 'output'

logger = make_logger(__name__)


def make_tag(tag_key: str, tag_value: int | str | list | set) -> str:
    if isinstance(tag_value, list) or isinstance(tag_value, set):
        if len(tag_value) == 0:
            raise ValueError("Tag value list/set cannot be empty")
        tag_value = ';'.join(str(x) for x in tag_value)
    elif not isinstance(tag_value, str):
        tag_value = str(tag_value)
    if len(tag_value) > 0:
        return f"{BWC_TAG_PREFIX}{tag_key}={tag_value}\n"
    else:
        return f"{BWC_TAG_PREFIX}{tag_key}\n"


## List tests files
def parse_excluded_tests(path):
    exclusion_list = {}
    with open(path) as f:
        for line in f:
            stripped = line.strip()
            if len(stripped) > 0 and line[0] != '#':
                exclusion_list[stripped] = True
    return exclusion_list


def find_tests_recursive(dir):
    test_list = []
    for f in os.listdir(dir):
        path = os.path.join(dir, f)
        if os.path.isdir(path):
            test_list += find_tests_recursive(path)
        elif path.endswith('.test'):
            test_list.append(path)
    return test_list


def list_test_files(root_dir: str, test_pattern: str = None) -> List[str]:
    test_dir = os.path.join(root_dir, 'test', 'sql')
    files = find_tests_recursive(test_dir)
    files.sort()
    files = files if test_pattern is None else [x for x in files if test_pattern in x]
    len_root_dir = len(root_dir) + 1
    return [f[len_root_dir:] for f in files]


class SerializedTest:
    class TestStep:
        def __init__(self, step_type: str, step_data: Any = None):
            self.step_type = step_type
            self.step_data = step_data

    def __init__(self, duckdb_bwc_base_dir: str, duckdb_version: str, test_filename: str, test_idx: int):
        # Provided metadata
        self.duckdb_bwc_base_dir = duckdb_bwc_base_dir
        self.duckdb_version = duckdb_version
        self.test_spec_relative_path = test_filename
        self.test_idx = test_idx

        # Computed paths
        test_basename = basename(test_filename)  # eg. attach_encryption_fallback_readonly.test
        if len(test_basename) == 0:
            raise ValueError(f"Test spec relative path '{test_filename}' is not valid (basename is empty)")

        self.test_specs_base_dir = f"{duckdb_bwc_base_dir}/specs/{duckdb_version}"

        is_spec_file = test_filename.startswith('test/sql/')
        if is_spec_file:
            self.test_absolute_filename = f"{duckdb_bwc_base_dir}/specs/{duckdb_version}/{test_filename}"
            self.test_runtime_directory = f"{duckdb_bwc_base_dir}/runtime/{duckdb_version}/{dirname(test_filename)}/{test_basename.replace('.test', '')}"
        else:
            self.test_absolute_filename = os.path.join(os.getcwd(), test_filename)
            self.test_runtime_directory = (
                f"{duckdb_bwc_base_dir}/runtime/{duckdb_version}/{test_basename.replace('.test', '')}"
            )

        self.test_output_directory = f"{self.test_runtime_directory}/{BWC_OUTPUT_DIR_NAME}"
        self.queries_file_name = f"{test_basename}.sql"

        self.serialized_plans_file_name = f"{test_basename}.plan.bin"
        self.results_old_file_name = f"{test_basename}.{duckdb_version}.result.bin"
        self.results_new_file_name = f"{test_basename}.new.result.bin"  # TODO - use actual version?

        # Log all paths
        logger.debug(f"SerializedTest#{self.test_idx} initialized:")
        logger.debug(f"  test_absolute_filename: {self.test_absolute_filename}")
        logger.debug(f"  test_runtime_directory: {self.test_runtime_directory}")
        logger.debug(f"  queries_file_name: {self.queries_file_name}")
        logger.debug(f"  serialized_plans_file_name: {self.serialized_plans_file_name}")
        logger.debug(f"  results_old_file_name: {self.results_old_file_name}")
        logger.debug(f"  results_new_file_name: {self.results_new_file_name}")

        # Used during parsing
        self.loaded_from_disk = False
        self.has_cached_serialized_plans = False
        self.ignore_error_messages = False
        self.no_extension_autoloading = False
        self.nb_skipped_statements_in_loop = 0
        self.steps = []

        # Used during execution
        self.skip_reasons = []
        self.fixtures = set()
        self.nb_steps = 0
        self.needed_extensions = set()

    def add(self, step_type: str, step_data: Any = None):
        self.steps.append(self.TestStep(step_type, step_data))

    def __str__(self):
        return (
            f"SerializedTest#{self.test_idx}(test_filename={self.test_absolute_filename}, steps={self.get_nb_steps()})"
        )

    def steps_str(self):
        return '\n- '.join([f"{step.step_type}: {step.step_data}" for step in self.steps])

    def create_runtime_directories(self):
        # Create runtime directory (base + output since they are nested)
        assert self.test_runtime_directory in self.test_output_directory
        os.makedirs(self.test_output_directory, exist_ok=True)

        # Create symlink to data directory so tests can access it via 'data/'
        data_symlink = f"{self.test_runtime_directory}/data"
        if not os.path.exists(data_symlink):
            os.symlink(f"{self.test_specs_base_dir}/data", data_symlink)

        test_symlink = f"{self.test_runtime_directory}/test"
        if not os.path.exists(test_symlink):
            os.symlink(f"{self.test_specs_base_dir}/test", test_symlink)

    def reset_output_directory(self):
        """
        When generating the SQL files, we also create some output fixtures (eg. during unzip)
        At the same time, during runtime the tests will also generate output files.
        So we before each run we copy all reference output directory.
        """
        output_files = self.list_output_directory(True)
        files_to_move = set()
        for output_file in output_files:
            if output_file in self.fixtures:
                continue

            # Only move the top level folder
            f = output_file.split('/')[0] if '/' in output_file else output_file
            files_to_move.add(f)

        # Ensure output directory exists
        os.makedirs(self.test_output_directory, exist_ok=True)

        # Restore fixtures
        for fixture in self.fixtures:
            shutil.copy2(f"{self.test_runtime_directory}/fixtures/{fixture}", self.test_output_directory)

        if len(files_to_move) == 0:
            return

        src = self.test_output_directory
        target_dir = f"{src}_{time.strftime('%Y%m%d_%H%M%S')}"

        # Target dir can exist if run with old version moved it just before
        if os.path.exists(target_dir):
            target_dir = f"{target_dir}_run2"

        os.mkdir(target_dir)
        for f in files_to_move:
            shutil.move(f"{src}/{f}", f"{target_dir}/{f}")

        logger.debug(f"Moved output directory from '{src}' to '{target_dir}'")

    def reload_from_disk(self) -> bool:
        """Reload a test from disk if it already exists."""
        absolute_queries_file = f"{self.test_runtime_directory}/{self.queries_file_name}"
        if not os.path.exists(absolute_queries_file):
            return False

        logger.debug(
            f"Reloading test '{self.test_absolute_filename}' from cached queries file '{absolute_queries_file}'"
        )
        with open(absolute_queries_file, 'r') as f:
            for full_line in f.readlines():
                line = full_line.strip()
                if not line.startswith(BWC_TAG_PREFIX):
                    continue

                [tag_key, _, tag_value] = line[len(BWC_TAG_PREFIX) :].partition('=')
                if tag_key == 'skip_reasons':
                    reasons = tag_value.split(';')
                    self.skip_reasons.extend(reasons)
                    break  # No need to parse further
                elif tag_key == 'needed_extensions':
                    extensions = tag_value.split(';')
                    self.needed_extensions.update(extensions)
                elif tag_key == 'nb_steps':
                    self.nb_steps = int(tag_value)
                elif tag_key == 'fixtures':
                    self.fixtures = set(tag_value.split(';'))

        self.loaded_from_disk = True
        self.has_cached_serialized_plans = self.exists_in_runtime_dir(
            self.serialized_plans_file_name
        ) and self.exists_in_runtime_dir(self.results_old_file_name)

        # Ensure symlinks exist (they are not included in the cache archive)
        self.create_runtime_directories()

        return True

    def exists_in_runtime_dir(self, file_name) -> bool:
        return os.path.exists(f"{self.test_runtime_directory}/{file_name}")

    def get_nb_steps(self) -> int:
        return self.nb_steps if self.loaded_from_disk else len(self.steps)

    def list_output_directory(self, allow_nested_path=False) -> List[str]:
        files = []
        folder = self.test_output_directory
        for root, _, filenames in os.walk(folder):
            for filename in filenames:
                f = os.path.relpath(os.path.join(root, filename), folder)
                if not allow_nested_path:
                    assert '/' not in f, f"Nested output files are not supported for now: {f}"
                files.append(f)

        return files

    def write_test_queries(self):
        if len(self.steps) == 0 and len(self.skip_reasons) == 0:
            logger.warning(f"No step found in test file '{self.queries_file_name}'")
            return

        gen_files = self.list_output_directory()
        self.fixtures.update(gen_files)

        if len(gen_files) > 0:
            os.mkdir(f"{self.test_runtime_directory}/fixtures")
            for gen_file in gen_files:
                shutil.copy2(f"{self.test_output_directory}/{gen_file}", f"{self.test_runtime_directory}/fixtures/")

        with open(f"{self.test_runtime_directory}/{self.queries_file_name}", 'w') as f:
            if len(self.skip_reasons) > 0:
                f.write(make_tag("skip_reasons", self.skip_reasons))
                return

            if len(gen_files) > 0:
                f.write(make_tag("fixtures", ';'.join(gen_files)))

            if len(self.needed_extensions) > 0:
                f.write(make_tag("needed_extensions", self.needed_extensions))

            f.write(make_tag("nb_steps", len(self.steps)))
            for step in self.steps:
                if step.step_type == "query":
                    f.write(step.step_data)
                else:
                    raise ValueError(f"Unsupported step type: {step.step_type}")


class SerializerSQLLogicContext(SQLLogicContext):
    def __init__(self, pool, runner, statements, keywords, update_value):
        super().__init__(pool, runner, statements, keywords, update_value)
        self.current_test = runner.current_test

    def update_settings(self):
        pass  # Nothing to do

    def _normalize_query(self, statement: Statement) -> str:
        # Normalize the query by removing comments and extra whitespace
        tags = None
        if statement.expected_result is not None and statement.expected_result.type != ExpectedResult.Type.SUCCESS:
            if statement.expected_result.type == ExpectedResult.Type.ERROR:
                tags = make_tag("expected_result", "error")
            elif statement.expected_result.type == ExpectedResult.Type.UNKNOWN:
                tags = make_tag("expected_result", "unknown")
            else:
                raise ValueError(f"Unknown expected result type: {statement.expected_result.type}")

        sql_query = '\n'.join(statement.lines)
        sql_query = self.replace_keywords(sql_query)
        return sql_query if not tags else f"{tags}\n{sql_query}"

    def execute_statement(self, statement: Statement):
        if len(statement.header.parameters) > 1:
            self.skiptest("Multiple connections not supported for now")
            return

        sql_query = self._normalize_query(statement)
        sql_query_lc = sql_query.lower()
        if 'set enable_external_access=false' in sql_query_lc:
            # Tester uses DuckDB API to write files...
            self.skiptest("Can't disable external access for now")
            return
        self.add_query(sql_query)

    def execute_query(self, query: Query):
        sql_query = ""

        # Handle sortstyle
        sort_style = query.get_sortstyle()
        if sort_style is not None and sort_style != SortStyle.NO_SORT:
            sql_query = make_tag("sort", SerializerSQLLogicContext.sort_style_to_str(sort_style))

        sql_query += self._normalize_query(query)

        self.add_query(sql_query)

    @staticmethod
    def sort_style_to_str(sort_style):
        if sort_style == SortStyle.NO_SORT:
            return "no_sort"
        elif sort_style == SortStyle.ROW_SORT:
            return "row_sort"
        elif sort_style == SortStyle.VALUE_SORT:
            return "value_sort"
        else:
            raise ValueError(f"Unknown sort style: {sort_style}")

    def add_query(self, query: str):
        if self.in_loop() and self.is_parallel:
            return  # Skip queries in parallel loops

        # Add exclusion tags based on query analysis
        exclusion_tags = format_exclusion_tags(query)
        if exclusion_tags:
            query = f"{exclusion_tags}\n{query}"

        self.current_test.add('query', f"{query}\n{make_tag('end_query', '')}\n")

    def execute_load(self, load: Load):
        options = []

        if not load.header.parameters:
            self.add_query(f"USE memory;")
            return

        dbpath = self.replace_keywords(load.header.parameters[0])
        db_name = basename(dbpath).split('.')[0]
        if load.readonly:
            options.append('READ_ONLY')
        if load.version:
            options.append(f"STORAGE_VERSION {load.version}")

        opt_str = f" ({', '.join(options)})" if options and len(options) > 1 else ''
        self.add_query(f"{make_tag('load_db', db_name)}\nATTACH '{dbpath}' AS {db_name}{opt_str};")

    def execute_restart(self, statement: Restart):
        # FIXME - skipping >200 tests because of this
        self.skiptest("Restart not supported for now")

    def execute_unzip(self, statement: Unzip):
        self.do_execute_unzip(statement)

    def execute_set(self, statement: Set):
        option = statement.header.parameters[0]
        if option == 'ignore_error_messages':
            self.current_test.ignore_error_messages = True
        elif option == 'seed':
            self.add_query(f"SELECT SETSEED({statement.header.parameters[1]});")
            # TODO - self.runner.skip_reload = True
        else:
            raise ValueError(f"SET '{option}' is not implemented!")

    def execute_foreach(self, foreach: Foreach):
        if not foreach.parallel:
            super().execute_foreach(foreach)
        else:
            # Get loop statements will skip over the inner statements
            skipped_stmts = self.get_loop_statements()
            self.current_test.nb_skipped_statements_in_loop += len(skipped_stmts)

    def execute_loop(self, loop: Loop):
        if not loop.parallel:
            super().execute_loop(loop)
        else:
            # Get loop statements will skip over the inner statements
            skipped_stmts = self.get_loop_statements()
            self.current_test.nb_skipped_statements_in_loop += len(skipped_stmts)

    def skiptest(self, reason):
        if reason not in self.current_test.skip_reasons:
            self.current_test.skip_reasons.append(reason)

    def check_require(self, statement: Require) -> RequireResult:
        not_supported = set(
            [
                "no_alternative_verify",
                "noalternativeverify",
                "fts",  # TODO - need to support non-linked extensions
            ]
        )
        not_an_extension = [
            # Copied over from scripts/sqllogictest/result.py
            "64bit",
            "block_size",
            "exact_vector_size",
            "longdouble",
            "mingw",
            "no_alternative_verify",
            "noalternativeverify",
            "noforcestorage",
            "nothreadsan",
            "notmingw",
            "notwindows",
            "skip_reload",
            "strinline",
            "vector_size",
            "windows",
            # Added here:
            "no_latest_storage",
            "no_vector_verification",
            "notmusl",
            "ram",
        ]

        param = statement.header.parameters[0].lower()
        if param in not_supported:
            # self.skiptest(f"Requirement '{param}' is not supported")
            return RequireResult.MISSING
        elif param in not_an_extension:
            # TODO - skip_reload?
            return RequireResult.PRESENT
        elif param == "no_extension_autoloading":
            # This is a special case, we do not want to load extensions automatically
            self.add_query("SET autoload_known_extensions=false;")
            self.current_test.no_extension_autoloading = True
            return RequireResult.PRESENT
        elif param == "allow_unsigned_extensions":
            # We're running with this parameter for now
            return RequireResult.PRESENT

        if not self.current_test.no_extension_autoloading:
            self.add_query(f"LOAD '{param}';")
            self.current_test.needed_extensions.add(param)
        return RequireResult.PRESENT


class SQLLogicTestSerializer(SQLLogicRunner):
    def __init__(self, duckdb_bwc_base_dir: str, duckdb_version: str, test_filename: str, test_idx: int):
        super().__init__(None)
        self.current_test = SerializedTest(duckdb_bwc_base_dir, duckdb_version, test_filename, test_idx)

    def execute_test(self) -> None:
        if self.current_test.reload_from_disk():
            return

        self.current_test.create_runtime_directories()

        # Set current working directory
        os.chdir(self.current_test.test_runtime_directory)

        self.reset()
        sql_parser = SQLLogicParser()
        test_absolute_filename = self.current_test.test_absolute_filename
        try:
            test = sql_parser.parse(str(test_absolute_filename))
            if test is None:
                raise ValueError(f"Failed to parse test file (unknown reason)")
        except Exception as e:
            # Throw if exception is file not found
            if not os.path.exists(test_absolute_filename):
                raise e

            self.current_test.skip_reasons.append("parsing_error")
            self.current_test.write_test_queries()
            logger.error(f"Exception while parsing test file '{test_absolute_filename}': {e}")
            return

        self.test = test
        self.extensions = [
            "httpfs",
            "parquet",
            "tpch",
            "json",
            "icu",
            "tpcds",
            "sqlite_scanner",
            "autocomplete",
            "inet",
        ]
        self.original_sqlite_test = self.test.is_sqlite_test()

        # Top level keywords
        keywords = {
            '__TEST_DIR__': BWC_OUTPUT_DIR_NAME,
            '__WORKING_DIRECTORY__': '.',
            '{DATA_DIR}': 'data',
            '{TEMP_DIR}': BWC_OUTPUT_DIR_NAME,
            '{TEST_DIR}': BWC_OUTPUT_DIR_NAME,
        }

        def update_value(_: SQLLogicContext) -> Generator[Any, Any, Any]:
            # Yield once to represent one iteration, do not touch the keywords
            yield None

        context = SerializerSQLLogicContext(None, self, test.statements, keywords, update_value)
        context.is_loop = False  # The outer context is not a loop
        context.verify_statements()
        res = context.execute()
        assert res.type == ExecuteResult.Type.SUCCESS

        if self.skip_active() and len(self.current_test.skip_reasons) == 0:
            self.current_test.skip_reasons.append("explicitly skipped")

        if (
            len(self.current_test.skip_reasons) == 0
            and len(self.current_test.steps) == 0
            and self.current_test.nb_skipped_statements_in_loop > 0
        ):
            self.current_test.skip_reasons.append("entirely_skipped_because_of_parallel_loops")

        self.current_test.write_test_queries()


def load_skipped_tests_from_file(json_file: str) -> set[str]:
    if not os.path.exists(json_file):
        return set()

    with open(json_file, 'r') as f:
        data = json.load(f)

    skipped_tests = set()
    for entry in data.get('skip_tests', []):
        skipped_tests.update(entry.get('paths', []))

    return skipped_tests


def list_skipped_tests(duckdb_test_config_dir: str, duckdb_version: str) -> set[str]:
    base_skip_file_path = f"{duckdb_test_config_dir}/test/configs/serialization_bwc_base.json"
    version_skip_file_path = f"{duckdb_test_config_dir}/test/configs/serialization_bwc_{duckdb_version}.json"
    skipped_tests = load_skipped_tests_from_file(base_skip_file_path)
    skipped_tests.update(load_skipped_tests_from_file(version_skip_file_path))
    logger.debug(
        f"Found {len(skipped_tests)} tests to skip in version {duckdb_version} (base skip file: {base_skip_file_path}, version skip file: {version_skip_file_path})"
    )
    return skipped_tests


class LoadingStats:
    def __init__(self):
        self.skipped_count = 0
        self.skipped_count_per_reason = {}
        self.start_time = time.time()
        self.nb_steps = 0
        self.nb_tests = 0
        self.nb_cached_queries = 0
        self.nb_cached_plans = 0

    def update(self, test: SerializedTest):
        if len(test.skip_reasons) > 0:
            self.update_skipped(test.skip_reasons)
            return

        self.nb_tests += 1
        self.nb_steps += test.get_nb_steps()

        if test.loaded_from_disk:
            self.nb_cached_queries += 1

        if test.has_cached_serialized_plans:
            self.nb_cached_plans += 1

    def update_skipped(self, reasons: List[str]):
        if len(reasons) == 0:
            return

        self.skipped_count += 1
        for reason in reasons:
            if reason not in self.skipped_count_per_reason:
                self.skipped_count_per_reason[reason] = 0
            self.skipped_count_per_reason[reason] += 1

    def log_stats(self):
        logger.info(
            f"Loaded {self.nb_tests} test files (with {self.nb_steps} steps) in {time.time() - self.start_time:.2f}s"
        )
        logger.info(
            f"  Cached queries: {self.nb_cached_queries} ({(self.nb_cached_queries/self.nb_tests*100) if self.nb_tests > 0 else 0:.2f}%)"
        )
        logger.info(
            f"  Cached plans: {self.nb_cached_plans} ({(self.nb_cached_plans/self.nb_tests*100) if self.nb_tests > 0 else 0:.2f}%)"
        )
        logger.info(
            f"  Skipped {self.skipped_count} ({(self.skipped_count/self.nb_tests*100) if self.nb_tests > 0 else 0:.2f}%) test files:"
        )

        # Sort self.skipped_count_per_reason by count descending
        sorted_skipped = sorted(self.skipped_count_per_reason.items(), key=lambda x: x[1], reverse=True)
        for reason, count in sorted_skipped:
            logger.info(
                f"    {count} ({(count/self.skipped_count*100) if self.skipped_count > 0 else 0:.2f}%): {reason}"
            )


def load_test_files(
    duckdb_root_dir: str,
    duckdb_bwc_base_dir: str,
    duckdb_version: str,
    test_pattern: str = None,
    single_test_file: str = None,
) -> List:
    loading_stats = LoadingStats()

    spec_files_dir = f"{duckdb_bwc_base_dir}/specs/{duckdb_version}"
    logger.info(f"Using Python DuckDB version {duckdb.__version__} to load test files from {spec_files_dir}")

    # Normalize (remove trailing ./ etc)
    single_test_file = os.path.normpath(single_test_file) if single_test_file is not None else None

    files = [single_test_file] if single_test_file is not None else list_test_files(spec_files_dir, test_pattern)
    logger.info(f"Found {len(files)} test files")

    loaded_tests = []
    skipped_tests = (
        list_skipped_tests(duckdb_root_dir, duckdb_version)
        if test_pattern is None and single_test_file is None
        else set()
    )
    needed_extensions = set()
    for test_filename in files:
        if test_filename in skipped_tests:
            loading_stats.update_skipped([f"in skipped file"])
            continue

        try:
            serializer = SQLLogicTestSerializer(
                duckdb_bwc_base_dir, duckdb_version, test_filename, len(loaded_tests) + 1
            )
            serializer.execute_test()
            test = serializer.current_test

            loading_stats.update(test)
            if len(test.skip_reasons) == 0:
                loaded_tests.append(test)
                needed_extensions.update(test.needed_extensions)
        except Exception as e:
            logger.error(f"Failed to load test file '{test_filename}'")
            raise e

    loading_stats.log_stats()

    return {'tests': loaded_tests, 'needed_extensions': needed_extensions}



import os
import argparse
import sys
import re
import subprocess
from typing import List, Dict
from pathlib import Path

SCRIPT_DIR = os.path.dirname(__file__)

parser = argparse.ArgumentParser(description="Generate a patch file for a DuckDB extension.")

parser.add_argument(
    "repository_path",
    type=str,
    help="Path to the repository where the changes live that should be turned into a patch.",
)

parser.add_argument(
    "extension_name",
    type=str,
    help="Name of the extension to patch, should match the name in `.github/config/extensions/<extension_name>.cmake`.",
)

parser.add_argument("patch_name", type=str, help="Name for the patch file to create.")

parser.add_argument("--overwrite", action="store_true", help="Allow overwriting the patch file if it already exists.")

args = parser.parse_args()


def verify_git_tag():
    # Locate the cmake file to extract the GIT_TAG from
    cmake_path = Path(SCRIPT_DIR) / '..' / ".github" / "config" / "extensions" / f"{args.extension_name}.cmake"
    if not cmake_path.is_file():
        print(f"Error: Extension CMake file not found: {cmake_path}")
        sys.exit(1)

    cmake_content = cmake_path.read_text()

    # Extract GIT_TAG from the cmake file
    match = re.search(r"\bGIT_TAG\s+([^\s\)]+)", cmake_content)
    if not match:
        print(f"Error: Could not find GIT_TAG in {cmake_path}")
        sys.exit(1)

    git_tag_in_cmake = match.group(1)

    # Get the current commit hash in repository_path
    try:
        result = subprocess.run(
            ["git", "rev-parse", "HEAD"],
            cwd=args.repository_path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=True,
        )
        current_commit = result.stdout.strip()
    except subprocess.CalledProcessError as e:
        print(f"Error: Failed to run git in {args.repository_path} — {e.stderr.strip()}")
        sys.exit(1)

    # Compare the tags
    if git_tag_in_cmake != current_commit:
        print(
            f"Error: GIT_TAG in {cmake_path} is {git_tag_in_cmake}, "
            f"but repository {args.repository_path} is checked out at {current_commit}."
        )
        sys.exit(1)


def create_patch():
    # Collect changes with git diff
    try:
        diff_result = subprocess.run(
            ["git", "diff", "--ignore-submodules"],
            cwd=args.repository_path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        print(f"Error: Failed to run git diff — {e.stderr.strip()}")
        sys.exit(1)

    new_patch_content = diff_result.stdout
    if not new_patch_content.strip():
        print("⚠️ No changes detected in repository; no patch will be created.")
        sys.exit(0)

    def parse_patch_files_and_lines(patch_text):
        changes = {}
        current_file = None
        for line in patch_text.splitlines():
            if line.startswith("diff --git"):
                parts = line.split()
                if len(parts) >= 3:
                    # Format: diff --git a/file b/file
                    current_file = parts[2][2:]  # remove 'a/'
                    changes.setdefault(current_file, set())
            elif line.startswith("@@") and current_file:
                # Format: @@ -old_start,old_count +new_start,new_count @@
                m = re.match(r"@@ -\d+(?:,\d+)? \+(\d+)(?:,(\d+))? @@", line)
                if m:
                    start = int(m.group(1))
                    length = int(m.group(2) or "1")
                    for l in range(start, start + length):
                        changes[current_file].add(l)
        return changes

    new_changes = parse_patch_files_and_lines(new_patch_content)

    # Check conflicts with existing patches
    patch_dir = (Path(SCRIPT_DIR) / ".." / ".github" / "patches" / "extensions" / args.extension_name).resolve()
    patch_dir.mkdir(parents=True, exist_ok=True)

    for existing_patch in patch_dir.glob("*.patch"):
        if existing_patch.name == f"{args.patch_name}.patch":
            if not args.overwrite:
                print(f"A patch by the name '{args.patch_name}.patch' already exists, failed to create patch")
                sys.exit(1)
            else:
                continue
        existing_changes = parse_patch_files_and_lines(existing_patch.read_text())

        for file, lines in new_changes.items():
            if file in existing_changes:
                overlap = lines & existing_changes[file]
                if overlap:
                    print(f"❌ Conflict detected with existing patch: {existing_patch.name}")
                    print(f"   File: {file}")
                    print(f"   Overlapping lines: {sorted(overlap)}")
                    sys.exit(1)

    # Save patch file
    patch_dir = (Path(SCRIPT_DIR) / ".." / ".github" / "patches" / "extensions" / args.extension_name).resolve()
    patch_dir.mkdir(parents=True, exist_ok=True)

    patch_path = patch_dir / f"{args.patch_name}.patch"
    patch_path.write_text(diff_result.stdout)


verify_git_tag()

create_patch()



# fmt: off

import pytest
import subprocess
import sys
from typing import List
from conftest import ShellTest
from conftest import json_extension
import os

@pytest.mark.skipif(os.name == 'nt', reason="Skipped on windows")
class TestReadFromStdin(object):

    def test_read_stdin_direct(self, shell):
        test = (
            ShellTest(shell)
            .input_file('data/csv/test/test.csv')
            .statement("""create table mytable as select * from read_text('/dev/stdin')""")
            .statement("select length(content) as len from mytable;")
            .add_argument(
                '-csv',
                ':memory:'
            )
        )
        result = test.run()
        result.check_stdout("len")
        result.check_stdout('77780')

    def test_read_stdin_csv(self, shell):
        test = (
            ShellTest(shell)
            .input_file('data/csv/test/test.csv')
            .statement("""
                create table mytable as select * from
                read_csv('/dev/stdin',
                    columns=STRUCT_PACK(foo := 'INTEGER', bar := 'INTEGER', baz := 'VARCHAR'),
                    AUTO_DETECT='false'
                )
            """)
            .statement("select * from mytable limit 1;")
            .add_argument(
                '-csv',
                ':memory:'
            )
        )
        result = test.run()
        result.check_stdout("foo,bar,baz")
        result.check_stdout('0,0, test')

    def test_read_stdin_csv_where_filename(self, shell):
        test = (
            ShellTest(shell)
            .input_file('data/csv/test/test.csv')
            .statement("""
                SELECT * FROM read_csv_auto(
                    'data/csv/bug_9005/teste*.csv',
                    header=TRUE,
                    filename=true,
                    union_by_name=True
                ) where filename='data/csv/bug_9005/teste1.csv'
            """)
            .add_argument(
                '-csv',
                ':memory:'
            )
        )
        result = test.run()
        result.check_stdout([
            'id,name,factor,filename',
            '1,Ricardo,1.5,data/csv/bug_9005/teste1.csv',
            '2,Jose,2.0,data/csv/bug_9005/teste1.csv'
        ])

    def test_read_stdin_csv_auto(self, shell):
        test = (
            ShellTest(shell)
            .input_file('data/csv/test/test.csv')
            .statement("""
                create table mytable as select * from
                read_csv_auto('/dev/stdin')
            """)
            .statement("select * from mytable limit 1;")
            .add_argument(
                '-csv',
                ':memory:'
            )
        )
        result = test.run()
        result.check_stdout("column0,column1,column2")
        result.check_stdout('0,0, test')

    def test_split_part_csv(self, shell):
        test = (
            ShellTest(shell)
            .input_file('data/csv/split_part.csv')
            .statement("""
                FROM read_csv('/dev/stdin') select split_part(C1, ',', 2) as res;
            """)
            .add_argument(
                '-csv',
                ':memory:'
            )
        )
        result = test.run()
        result.check_stdout("res")
        result.check_stdout('12')
        result.check_stdout('12')

    def test_read_stdin_csv_auto_projection(self, shell):
        test = (
            ShellTest(shell)
            .input_file('data/csv/tpcds_14.csv')
            .statement("""
                create table mytable as select * from
                read_csv_auto('/dev/stdin')
            """)
            .statement("select channel,i_brand_id,sum_sales,number_sales from mytable;")
            .add_argument(
                '-csv',
                ':memory:'
            )
        )
        result = test.run()
        result.check_stdout("web,8006004,844.21,21")


    def test_read_stdin_ndjson(self, shell, json_extension):
        test = (
            ShellTest(shell)
            .input_file('data/json/example_rn.ndjson')
            .statement("""
                create table mytable as select * from
                read_ndjson_objects('/dev/stdin')
            """)
            .statement("select * from mytable;")
            .add_argument(
                '-list',
                ':memory:'
            )
        )
        result = test.run()
        result.check_stdout([
            "json",
            '{"id":1,"name":"O Brother, Where Art Thou?"}',
            '{"id":2,"name":"Home for the Holidays"}',
            '{"id":3,"name":"The Firm"}',
            '{"id":4,"name":"Broadcast News"}',
            '{"id":5,"name":"Raising Arizona"}'
        ])


    def test_read_stdin_json_auto(self, shell, json_extension):
        test = (
            ShellTest(shell)
            .input_file('data/json/example_rn.ndjson')
            .statement("""
                create table mytable as select * from
                read_json_auto('/dev/stdin')
            """)
            .statement("select * from mytable;")
            .add_argument(
                '-list',
                ':memory:'
            )
        )
        result = test.run()
        result.check_stdout([
            'id|name',
            '1|O Brother, Where Art Thou?',
            '2|Home for the Holidays',
            '3|The Firm',
            '4|Broadcast News',
            '5|Raising Arizona'
        ])

    def test_read_stdin_json_array(self, shell, json_extension):
        test = (
            ShellTest(shell)
            .input_file('data/json/11407.json')
            .statement("""
                create table mytable as select * from
                read_json_auto('/dev/stdin')
            """)
            .statement("select * from mytable;")
            .add_argument(
                '-list',
                ':memory:'
            )
        )
        result = test.run()
        result.check_stdout([
            'k',
            'v',
            'v2'
        ])

    def test_read_stdin_json_auto_recursive_cte(self, shell, json_extension):
        # FIXME: disabled for now
        return
        test = (
            ShellTest(shell)
            .input_file('data/json/filter_keystage.ndjson')
            .statement("""
                CREATE TABLE mytable AS
                WITH RECURSIVE nums AS (
                    SELECT 0 AS n
                UNION ALL
                    SELECT n + 1
                    FROM nums
                    WHERE n < (
                        SELECT MAX(JSON_ARRAY_LENGTH(filter_keystage))::int - 1
                        FROM read_json_auto('/dev/stdin'))
                ) SELECT * FROM nums;
            """)
            .statement("select * from mytable;")
            .add_argument(
                '-list',
                ':memory:'
            )
        )
        result = test.run()
        result.check_stdout([
            'n',
            '0',
            '1',
            '2',
        ])


    @pytest.mark.parametrize("alias", [
        "'/dev/stdout'",
        'stdout'
    ])
    def test_copy_to_stdout(self, shell, alias):
        test = (
            ShellTest(shell)
            .statement(f"COPY (SELECT 42) TO {alias};")
        )
        result = test.run()
        result.check_stdout('42')


    @pytest.mark.parametrize("alias", [
        "'/dev/stdout'",
        'stdout'
    ])
    def test_copy_csv_to_stdout(self, shell, alias):
        test = (
            ShellTest(shell)
            .statement(f"COPY (SELECT 42) TO {alias} WITH (FORMAT 'csv');")
            .add_argument(
                '-csv',
                ':memory:'
            )
        )
        result = test.run()
        result.check_stdout('42')


    @pytest.mark.parametrize("alias", [
        "'/dev/stderr'"
    ])
    def test_copy_csv_to_stderr(self, shell, alias):
        test = (
            ShellTest(shell)
            .statement(f"COPY (SELECT 42) TO {alias} WITH (FORMAT 'csv');")
            .add_argument(
                '-csv',
                ':memory:'
            )
        )
        result = test.run()
        result.check_stderr('42')


    def test_copy_non_inlined_string(self, shell):
        test = (
            ShellTest(shell)
            .statement("select list(concat('thisisalongstring', range::VARCHAR)) i from range(10000)")
        )
        result = test.run()
        result.check_stdout('thisisalongstring')


    def test_write_to_stdout_piped_to_file(self, shell, random_filepath):
        test = (
            ShellTest(shell)
            .statement("copy (select * from range(10000) tbl(i)) to '/dev/stdout' (format csv)")
            .output_file(random_filepath.as_posix())
        )
        result = test.run()
        result.check_stdout('9999')

# fmt: on



# fmt: off

import pytest
import subprocess
import sys
from typing import List
from conftest import ShellTest
import os
from pathlib import Path


# test import from a parquet file
def test_import_parquet(shell):
    test = (
        ShellTest(shell)
        .statement(f'.import data/parquet-testing/unsigned.parquet a')
        .statement("SELECT MAX(d) FROM a")
    )

    result = test.run()
    result.check_stdout("18446744073709551615")

# test import from a json file
def test_import_json(shell):
    test = (
        ShellTest(shell)
        .statement(f'.import data/json/example_n.ndjson a')
        .statement("SELECT name FROM a")
    )

    result = test.run()
    result.check_stdout("Broadcast News")


# fmt: on



def open_utf8(fpath, flags, **kwargs):
    import sys

    if sys.version_info[0] < 3:
        return open(fpath, flags, **kwargs)
    else:
        return open(fpath, flags, encoding="utf8", **kwargs)


def normalize_path(path):
    import os

    def normalize(p):
        return os.path.sep.join(p.split('/'))

    if isinstance(path, list):
        normed = map(lambda p: normalize(p), path)
        return list(normed)

    if isinstance(path, str):
        return normalize(path)

    raise Exception("Can only be called with a str or list argument")



from metrics.emit_enum_cpp import generate_metric_type_files
from metrics.emit_profiling_utils_cpp import generate_profiling_utils
from metrics.emit_tests import generate_test_files
from metrics.inputs import load_metrics_json, retrieve_optimizers
from metrics.model import build_all_metrics
from metrics.paths import (
    METRICS_JSON,
    OPTIMIZER_HPP,
    OUT_METRIC_HPP,
    OUT_METRIC_CPP,
    TEST_PROFILING_DIR,
    OUT_PROFILING_HPP,
    OUT_PROFILING_CPP,
    path_from_duckdb,
    format_file,
)

if __name__ == "__main__":
    # load metrics from JSON
    metrics_json = load_metrics_json(METRICS_JSON)

    optimizers = retrieve_optimizers(OPTIMIZER_HPP)

    metric_index = build_all_metrics(metrics_json, optimizers)

    print(f"Metric Information:")
    print(f"  * Total Metrics: {len(metric_index.metrics_by_group['all'])}")
    print(f"  * Default Metrics: {len(metric_index.metrics_by_group['default'])}")
    print(f"  * Total Groups: {len(metric_index.group_names)}")
    print(f"  * Metric Groups and Total Metrics per Group:")
    for group in metric_index.group_names:
        if group == "all" or group == "default":
            continue
        print(f"    * {group}: {len(metric_index.metrics_by_group[group])}")

    print("\nGenerating files:")
    # emit C++ files
    generate_metric_type_files(OUT_METRIC_HPP, OUT_METRIC_CPP, metric_index, optimizers)
    print(f"  * {path_from_duckdb(OUT_METRIC_HPP)}")
    format_file(OUT_METRIC_HPP)
    print(f"  * {path_from_duckdb(OUT_METRIC_CPP)}")
    format_file(OUT_METRIC_CPP)

    generate_profiling_utils(OUT_PROFILING_HPP, OUT_PROFILING_CPP, metric_index)
    print(f"  * {path_from_duckdb(OUT_PROFILING_HPP)}")
    format_file(OUT_PROFILING_HPP)
    print(f"  * {path_from_duckdb(OUT_PROFILING_CPP)}")
    format_file(OUT_PROFILING_CPP)

    print("\nGenerating tests:")
    # emit test files
    generate_test_files(TEST_PROFILING_DIR, metric_index.metrics_by_group)



import json
from .config import Setting, SettingsList, JSON_PATH


# sort settings in json by name
def sort_json_data(path):
    with open(path, 'r') as file:
        data = json.load(file)
    sorted_data = sorted(data, key=lambda x: x['name'])
    with open(path, 'w') as file:
        json.dump(sorted_data, file, indent=4)
    return sorted_data


# parse json data and stores each entry as a settings object in the global list SettingsList
def add_all_settings_to_global_list():
    valid_entries = [
        'name',
        'description',
        'type',
        'scope',
        'internal_setting',
        'on_callbacks',
        'custom_implementation',
        'struct',
        'aliases',
        'default_scope',
        'default_value',
        'conditional_defaults',
    ]

    print(f"Parsing and sorting the settings data in {JSON_PATH}")
    clear_global_settings_list()
    json_data = sort_json_data(JSON_PATH)
    # store all the settings in the SettingsList
    for entry in json_data:
        for field_entry in entry:
            if field_entry not in valid_entries:
                raise ValueError(
                    f"Found entry unexpected entry \"{field_entry}\" in setting, expected entry to be in {', '.join(valid_entries)}"
                )
        setting = Setting(
            name=entry['name'],
            description=entry['description'],
            sql_type=entry['type'],
            internal_setting=entry.get('internal_setting', entry['name']),
            scope=entry.get('scope', None),
            struct_name=entry.get('struct', ''),
            on_callbacks=entry.get('on_callbacks', []),
            custom_implementation=entry.get('custom_implementation', False),
            aliases=entry.get('aliases', []),
            default_scope=entry.get('default_scope', None),
            default_value=entry.get('default_value', None),
            conditional_defaults=entry.get('conditional_defaults', None),
        )
        SettingsList.append(setting)


def clear_global_settings_list():
    SettingsList.clear()



import os
import subprocess
import tempfile
from pathlib import Path
import urllib.request
from utils.logger import make_logger

from .duckdb_cli import DuckDBCLI

logger = make_logger(__name__)


def make_cli_path(version):
    raw_version = version.lstrip('v')
    cli_path = Path.home() / ".duckdb" / "cli" / raw_version / "duckdb"
    return str(cli_path)


def install_duckdb_cli(version):
    # Check if the cli already exists
    cli_path = make_cli_path(version)
    if os.path.exists(cli_path):
        logger.debug(f"DuckDB CLI {version} already exists at '{cli_path}'")
        return

    req = urllib.request.Request("https://install.duckdb.org", headers={"User-Agent": "duckdb-bwc-test-runner"})
    with urllib.request.urlopen(req) as response:
        script_content = response.read()

    env = os.environ.copy()
    env['DUCKDB_VERSION'] = version.lstrip('v')

    result = subprocess.run(
        ['sh'],
        input=script_content,
        env=env,
        capture_output=True,
        text=False,  # Keep as bytes since we're reading bytes
    )

    if result.returncode != 0:
        raise RuntimeError(f"[{version}] Install script failed: {result.stderr.decode()}")

    if not os.path.exists(cli_path):
        raise RuntimeError(
            f"[{version}] Install script succeeded but CLI not found at '{cli_path}'. stdout: {result.stdout.decode()}"
        )

    logger.info(f"[{version}] DuckDB CLI installed successfully at '{cli_path}'")


def download_test_specs(version, target_sub_dir):
    logger.debug(f"[{version}] Downloading DuckDB test specs")
    target_dir = os.path.join(target_sub_dir, version)
    if os.path.exists(target_dir):
        logger.info(f"[{version}] Test specs already exist at '{target_dir}'")
        return

    with tempfile.TemporaryDirectory() as tmpdirname:
        logger.info(f"[{version}] Cloning duckdb repo in {tmpdirname}")
        subprocess.run(
            ["git", "clone", "--depth", "1", "--branch", version, f"https://github.com/duckdb/duckdb.git"],
            cwd=tmpdirname,
            check=True,
            capture_output=True,
            text=True,
        )
        logger.info(f"[{version}] Cloned duckdb repo.")

        # Move `test` and `data` directories to target_dir
        os.makedirs(target_dir)
        subprocess.run(["mv", f"duckdb/test", target_dir], cwd=tmpdirname, check=True)
        subprocess.run(["mv", f"duckdb/data", target_dir], cwd=tmpdirname, check=True)
        logger.info(f"[{version}] Copied test specs to '{target_dir}'")


def get_version(cli_path):
    with DuckDBCLI(cli_path) as cli:
        result = cli.execute_command("select library_version, source_id from pragma_version();")
        if not result["success"]:
            raise RuntimeError(f"Failed to get DuckDB version from CLI at '{cli_path}': {result['stderr']}")
        return result["output"][1].split(',')


def install_extensions(cli_path, extensions):
    with DuckDBCLI(cli_path, unsigned=True) as cli:
        installed = []
        failed = []

        # Happy path - all extensions installed
        ext_list = ", ".join(f"'{ext}'" for ext in extensions)
        result = cli.execute_command(
            f"SELECT COUNT(*) FROM duckdb_extensions() WHERE extension_name in ({ext_list}) AND installed;"
        )
        if result["output"][1] == str(len(extensions)):
            logger.info(f"[{cli.version}] All {len(extensions)} extensions were already installed")
            return

        for ext in extensions:
            # First look if it's already installed
            result = cli.execute_command(
                f"SELECT COUNT(*) FROM duckdb_extensions() WHERE extension_name='{ext}' AND installed;"
            )
            if result["output"][1] == '1':
                logger.debug(f"[{cli.version}] Extension '{ext}' is already installed.")
                installed.append(ext)
                continue

            if ext == 'test_utils':
                result = cli.execute_command(
                    f"INSTALL 'test_utils' FROM 'https://raw.githubusercontent.com/duckdb/bwc-test-utils/main/extensions';"
                )
            else:
                result = cli.execute_command(f"INSTALL '{ext}'")

            if result["success"]:
                installed.append(ext)
            else:
                failed.append(ext)

        logger.info(f"[{cli.version}] Installed {len(installed)} extensions.")

        if len(failed):
            root_ext_cmake_path = Path(__file__).parent.parent.parent.parent / ".github/config/extensions"
            if not root_ext_cmake_path.is_dir():
                raise RuntimeError(
                    f"Extensions CMake config directory does not exist at '{root_ext_cmake_path}' - did the repo structure change?"
                )

            # test if `.github/config/extensions/{extension}.cmake` is a file
            cmake_files_ext = []
            core_extensions = []
            for ext in failed:
                ext_filename = "test-utils" if ext == "test_utils" else ext
                ext_cmake_path = root_ext_cmake_path / f"{ext_filename}.cmake"
                if ext_cmake_path.is_file():
                    cmake_files_ext.append(ext_filename)
                else:
                    core_extensions.append(ext)

            # Build a string that suggests how to install the failed extensions
            install_cmd = f"make release "
            if len(cmake_files_ext) > 0:
                install_cmd += (
                    "EXTENSION_CONFIGS='"
                    + ";".join([f".github/config/extensions/{ext}.cmake" for ext in cmake_files_ext])
                    + "' "
                )
            if len(core_extensions) > 0:
                install_cmd += "DUCKDB_EXTENSIONS='" + ";".join(core_extensions) + "'"

            if len(cmake_files_ext) > 0 or len(core_extensions) > 0:
                logger.info(
                    f"[{cli.version}] Failed to install {len(failed)} extensions. If you need to compile them:\n{install_cmd}"
                )
            raise RuntimeError(f"[{cli.version}] Failed to install {len(failed)} extensions: {', '.join(failed)}")


def install_assets(version, bwc_tests_base_dir):
    install_duckdb_cli(version)
    download_test_specs(version, f"{bwc_tests_base_dir}/specs")



import duckdb

# basic SQL API

# connect to an in-memory temporary database
conn = duckdb.connect()

# if you want, you can create a cursor() like described in PEP 249 but it's fully redundant
cursor = conn.cursor()

# run arbitrary SQL commands
conn.execute("CREATE TABLE test_table (i INTEGER, j STRING)")

# add some data
conn.execute("INSERT INTO test_table VALUES (1, 'one')")

# we can use placeholders for parameters
conn.execute("INSERT INTO test_table VALUES (?, ?)", [2, 'two'])

# we can provide multiple sets of parameters to executemany()
conn.executemany("INSERT INTO test_table VALUES (?, ?)", [[3, 'three'], [4, 'four']])

# fetch as pandas data frame
print(conn.execute("SELECT * FROM test_table").fetchdf())

# fetch as list of masked numpy arrays, cleaner when handling NULLs
print(conn.execute("SELECT * FROM test_table").fetchnumpy())


# we can query pandas data frames as if they were SQL views
# create a sample pandas data frame
import pandas as pd

test_df = pd.DataFrame.from_dict({"i": [1, 2, 3, 4], "j": ["one", "two", "three", "four"]})

# make this data frame available as a view in duckdb
conn.register("test_df", test_df)
print(conn.execute("SELECT j FROM test_df WHERE i > 1").fetchdf())


# relation API, programmatic querying. relations are lazily evaluated chains of relational operators

# create a "relation" from a pandas data frame with an existing connection
rel = conn.from_df(test_df)
print(rel)

# alternative shorthand, use a built-in default connection to create a relation from a pandas data frame
rel = duckdb.df(test_df)
print(rel)

# create a relation from a CSV file

# first create a CSV file from our pandas example
import tempfile, os

temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
test_df.to_csv(temp_file_name, index=False)

# now create a relation from it
rel = duckdb.from_csv_auto(temp_file_name)
print(rel)

# create a relation from an existing table
rel = conn.table("test_table")
print(rel)

# a relation has an alias (like a table name)
print(rel.alias)

# we can change the alias, useful for (self)joins for example
rel2 = rel.set_alias('new_alias')
print(rel2.alias)

# we can inspect the type of a relation
print(rel.type)

# or the column names that are in it
print(rel.columns)

# or the types of those columns
print(rel.types)

# now we can apply some operators to the relation
# filter the relation
print(rel.filter('i > 1'))

# project the relation, get some columns
print(rel.project('i, j'))

# or transform them
print(rel.project('i + 1'))

# order the relation
print(rel.order('j'))

# limit the rows returned
print(rel.limit(2))

# skip the first row and limit the number of results
print(rel.limit(2, offset=1))

# of course these things can be chained
print(rel.filter('i > 1').project('i + 1, j').order('j').limit(2))

# aggregate the relation
print(rel.aggregate("sum(i)"))

# non-aggregated columns create implicit grouping
print(rel.aggregate("j, sum(i)"))

# we can also explicit group the relation before aggregating
print(rel.aggregate("sum(i)", "j"))

# distinct values
print(rel.distinct())


# multi-relation operators are also supported, e.g union
print(rel.union(rel))

# join rel with itself on i
rel2 = conn.from_df(test_df)
print(rel.join(rel2, 'i'))

# for explicit join conditions the relations can be named using alias()
print(rel.set_alias('a').join(rel.set_alias('b'), 'a.i=b.i'))


# there are also shorthand methods to directly create a relation and apply an operator from pandas data frame objects
print(duckdb.filter(test_df, 'i > 1'))
print(duckdb.project(test_df, 'i + 1'))
print(duckdb.order(test_df, 'j'))
print(duckdb.limit(test_df, 2))

print(duckdb.aggregate(test_df, "sum(i)"))
print(duckdb.distinct(test_df))

# when chaining only the first call needs to include the data frame parameter
print(duckdb.filter(test_df, 'i > 1').project('i + 1, j').order('j').limit(2))

# turn the relation into something else again


# compute the query result from the relation
res = rel.execute()
print(res)
# res is a query result, you can call fetchdf() or fetchnumpy() or fetchone() on it
print(res.fetchone())
print(res.fetchall())

# convert a relation back to a pandas data frame
print(rel.to_df())

# df() is shorthand for to_df() on relations
print(rel.df())

# create a table in duckdb from the relation
print(rel.create("test_table2"))

# insert the relation's data into an existing table
conn.execute("CREATE TABLE test_table3 (i INTEGER, j STRING)")
print(rel.insert_into("test_table3"))

# Inserting elements into table_3
print(conn.values([5, 'five']).insert_into("test_table3"))
rel_3 = conn.table("test_table3")
rel_3.insert([6, 'six'])

# create a SQL-accessible view of the relation
print(rel.create_view('test_view'))


# we can also directly run SQL queries on relation objects without explicitly creating a view
# the first parameter gives the rel object a view name so we can refer to it in queries
res = rel.query('my_name_for_rel', 'SELECT * FROM my_name_for_rel')
print(res)
# res is a query result, we can fetch with the methods described above, e.g.
print(res.fetchone())
print(res.fetchdf())
# or just use df(), a shorthand for fetchdf() on query results
print(res.df())

# this also works directly on data frames
res = duckdb.query_df(test_df, 'my_name_for_test_df', 'SELECT * FROM my_name_for_test_df')
print(res.df())



# scripts/metrics/emit_tests.py
from __future__ import annotations

from pathlib import Path
from typing import Dict, List

from .paths import REPO_ROOT, path_from_duckdb, format_file
from .writer import IndentedFileWriter


def _write_statement(f, statement_type, statement):
    f.write(f"statement {statement_type}\n")
    f.write(statement + "\n\n")


def _write_query(f, options, query):
    f.write(f"query {options}\n")
    f.write(query + "\n")
    f.write("----\n")


def _write_default_query(f):
    query = "SELECT unnest(['Maia', 'Thijs', 'Mark', 'Hannes', 'Tom', 'Max', 'Carlo', 'Sam', 'Tania']) AS names ORDER BY random();"
    _write_statement(f, "ok", query)
    _write_statement(f, "ok", "PRAGMA disable_profiling;")


def _write_get_custom_profiling_settings(f):
    query = """
SELECT unnest(res) FROM (
    SELECT current_setting('custom_profiling_settings') AS raw_setting,
    raw_setting.trim('{}') AS setting,
    string_split(setting, ', ') AS res
) ORDER BY ALL;
""".strip()
    _write_query(f, "I", query)


def _write_custom_profiling_optimizer(f):
    _write_statement(f, "ok", "PRAGMA custom_profiling_settings='{\"ALL_OPTIMIZERS\": \"true\"}';")

    _write_default_query(f)

    query = """
SELECT * FROM (
    SELECT unnest(res) str FROM (
        SELECT current_setting('custom_profiling_settings') as raw_setting,
        raw_setting.trim('{}') AS setting,
        string_split(setting, ', ') AS res
    )
) WHERE '"true"' NOT in str
ORDER BY ALL
""".strip()
    _write_query(f, "I", query)
    f.write("\n")

    _write_statement(f, "ok", "PRAGMA custom_profiling_settings='{}'")
    _write_default_query(f)

    _write_get_custom_profiling_settings(f)
    f.write("(empty)\n\n")

    _write_statement(f, "ok", "PRAGMA custom_profiling_settings='{\"OPTIMIZER_JOIN_ORDER\": \"true\"}'")
    _write_default_query(f)

    _write_get_custom_profiling_settings(f)
    f.write("\"OPTIMIZER_JOIN_ORDER\": \"true\"\n\n")

    _write_statement(
        f, "ok", "CREATE OR REPLACE TABLE metrics_output AS SELECT * FROM '__TEST_DIR__/profiling_output.json';"
    )

    query = """
SELECT
    CASE WHEN optimizer_join_order > 0 THEN 'true'
     ELSE 'false' END
FROM metrics_output;
""".strip()
    _write_query(f, "I", query)
    f.write("true\n\n")

    _write_statement(f, "ok", "SET disabled_optimizers = 'JOIN_ORDER';")
    _write_statement(f, "ok", "PRAGMA custom_profiling_settings='{\"OPTIMIZER_JOIN_ORDER\": \"true\"}'")
    _write_default_query(f)

    _write_get_custom_profiling_settings(f)
    f.write("(empty)\n\n")

    _write_statement(f, "ok", "PRAGMA custom_profiling_settings='{\"CUMULATIVE_OPTIMIZER_TIMING\": \"true\"}';")
    _write_default_query(f)

    _write_statement(
        f, "ok", "CREATE OR REPLACE TABLE metrics_output AS SELECT * FROM '__TEST_DIR__/profiling_output.json';"
    )

    query = """
SELECT
    CASE WHEN cumulative_optimizer_timing > 0 THEN 'true'
    ELSE 'false' END
FROM metrics_output;
""".strip()
    _write_query(f, "I", query)
    f.write("true\n\n")

    f.write("# All phase timings must be collected when using detailed profiling mode.\n\n")

    _write_statement(f, "ok", "RESET custom_profiling_settings;")
    _write_statement(f, "ok", "SET profiling_mode = 'detailed';")
    _write_default_query(f)

    query = """
SELECT * FROM (
    SELECT unnest(res) str FROM (
        SELECT current_setting('custom_profiling_settings') AS raw_setting,
        raw_setting.trim('{}') AS setting,
        string_split(setting, ', ') AS res
    )
)
WHERE '"true"' NOT IN str
ORDER BY ALL
""".strip()
    _write_query(f, "I", query)
    f.write("\n")

    _write_statement(f, "ok", "RESET custom_profiling_settings;")
    _write_statement(f, "ok", "SET profiling_mode = 'standard';")


def _generate_group_test(f, groups: list[str], all_metrics: Dict[str, List[str]]):
    _write_statement(f, "ok", "PRAGMA enable_profiling = 'json';")
    _write_statement(f, "ok", "PRAGMA profiling_output = '__TEST_DIR__/profiling_output.json';")

    group_str = ", ".join(f'"{g.upper()}": "true"' for g in groups)
    _write_statement(f, "ok", f"PRAGMA custom_profiling_settings='{{{group_str}}}';")

    _write_default_query(f)
    _write_get_custom_profiling_settings(f)

    metrics: list[str] = []
    for g in groups:
        metrics += all_metrics[g]

    if "all" not in groups and "ALL_OPTIMIZERS" in metrics:
        metrics.extend(all_metrics.get("optimizer", []))

    metrics = list(set(metrics))
    metrics.sort()
    for m in metrics:
        f.write(f'"{m}": "true"\n')
    f.write("\n")

    _write_statement(
        f, "ok", "CREATE OR REPLACE TABLE metrics_output AS SELECT * FROM '__TEST_DIR__/profiling_output.json';"
    )

    cols: list[str] = []
    operator_metrics = set(all_metrics.get("operator", []))
    for m in metrics:
        if m in operator_metrics and "operator" not in groups:
            continue
        cols.append(m)

    select = "SELECT " + ",\n\t".join(cols)

    if "operator" in groups:
        select += "\nFROM (\n"
        select += "\tSELECT unnest(children, max_depth := 2)\n"
        select += "\tFROM metrics_output\n"
        select += ")"
    else:
        select += "\nFROM metrics_output;"

    _write_statement(f, "ok", select)


def _generate_metric_group_test_file(out_path, all_metrics: Dict[str, List[str]]):
    name = path_from_duckdb(out_path)
    print(f"  * {name}")

    top = f"""# name: {name}
# description: Test default profiling settings using groups.
# group: [profiling]

# This file is automatically generated by scripts/generate_metric_enums.py
# Do not edit this file manually, your changes will be overwritten

require json

"""
    with IndentedFileWriter(out_path) as f:
        f.write(top)
        for group in all_metrics:
            _generate_group_test(f, [group], all_metrics)
        _generate_group_test(f, ["default", "file"], all_metrics)
        _generate_group_test(f, ["file", "optimizer"], all_metrics)
        _generate_group_test(f, ["phase_timing", "execution", "file"], all_metrics)
    format_file(out_path)


def _generate_profiling_setting_tests(out_dir: Path, all_metrics: Dict[str, List[str]]):
    test_names = [
        "test_default_profiling_settings",
        "test_custom_profiling_optimizer_settings",
        "test_all_profiling_settings",
    ]
    test_descriptions = ["default", "custom optimizer", "all settings"]
    test_paths = [out_dir / f"{name}.test" for name in test_names]
    metrics_group = ["default", "default", "all"]

    for test_file, name, description, group in zip(test_paths, test_names, test_descriptions, metrics_group):
        display_name = path_from_duckdb(test_file)
        print(f"  * {display_name}")
        with IndentedFileWriter(test_file) as f:
            f.write(f"# name: {display_name}\n")
            f.write(f"# description: Test {description} profiling settings.\n")
            f.write("# group: [profiling]\n\n")
            f.write("# This file is automatically generated by scripts/generate_metric_enums.py\n")
            f.write("# Do not edit this file manually, your changes will be overwritten\n\n")
            f.write("require json\n\n")

            _write_statement(f, "ok", "PRAGMA enable_profiling = 'json';")
            _write_statement(f, "ok", "PRAGMA profiling_output = '__TEST_DIR__/profiling_output.json';")

            mode = "standard" if group == "default" else group
            _write_statement(f, "ok", f"SET profiling_mode='{mode}';")

            if name == "test_custom_profiling_optimizer_settings":
                _write_custom_profiling_optimizer(f)

            _write_default_query(f)
            _write_get_custom_profiling_settings(f)

            for m in all_metrics[group]:
                f.write(f'"{m}": "true"\n')
            f.write("\n")

            _write_statement(
                f, "ok", "CREATE OR REPLACE TABLE metrics_output AS SELECT * FROM '__TEST_DIR__/profiling_output.json';"
            )
            _write_statement(f, "ok", "SELECT cpu_time, extra_info, rows_returned, latency FROM metrics_output;")
        format_file(test_file)


def generate_test_files(out_dir: Path, all_metrics: Dict[str, List[str]]):
    _generate_profiling_setting_tests(out_dir, all_metrics)
    _generate_metric_group_test_file(out_dir / "test_custom_profiling_using_groups.test", all_metrics)



# fmt: off

from conftest import ShellTest

def test_temp_directory(shell):
    test = (
        ShellTest(shell)
        .statement(".mode csv")
        .statement("CREATE SEQUENCE id_seq;")
        .statement("""
            CREATE TABLE my_table (
            id INTEGER DEFAULT nextval('id_seq'),
            a INTEGER
        );""")
        .statement("ATTACH ':memory:' AS s1;")
        .statement("CREATE TABLE s1.tbl AS FROM range(2000000);")
        .statement("INSERT INTO my_table (a) SELECT * FROM s1.tbl;")
        .statement("INSERT INTO my_table (a) SELECT * FROM s1.tbl;")
        .statement("INSERT INTO my_table (a) SELECT * FROM s1.tbl;")
        .statement("INSERT INTO my_table (a) SELECT * FROM s1.tbl;")
        .statement("INSERT INTO my_table (a) SELECT * FROM s1.tbl;")
        .statement("INSERT INTO my_table (a) SELECT * FROM s1.tbl;")
        .statement("INSERT INTO my_table (a) SELECT * FROM s1.tbl;")
        .statement("INSERT INTO my_table (a) SELECT * FROM s1.tbl;")
        .statement("INSERT INTO my_table (a) SELECT * FROM s1.tbl;")
        .statement("INSERT INTO my_table (a) SELECT * FROM s1.tbl;")
    )
    test = test.statement("select count(*) from my_table")
    result = test.run()
    result.check_stdout("20000000")

# fmt: on



from __future__ import annotations

from typing import Dict, List, Tuple
from pathlib import Path

from .inputs import _to_pascal_case
from .model import MetricIndex
from .writer import IndentedFileWriter, write_warning

HPP_HEADER = """

#pragma once

#include "duckdb/common/types/value.hpp"
#include "duckdb/common/unordered_set.hpp"
#include "duckdb/common/constants.hpp"
#include "duckdb/common/enums/optimizer_type.hpp"

namespace duckdb {

"""

HPP_TYPEDEFS = """
struct MetricTypeHashFunction {
    uint64_t operator()(const MetricType &index) const {
        return std::hash<uint8_t>()(static_cast<uint8_t>(index));
    }
};

typedef unordered_set<MetricType, MetricTypeHashFunction> profiler_settings_t;
typedef unordered_map<MetricType, Value, MetricTypeHashFunction> profiler_metrics_t;

"""

CPP_HEADER = """
#include "duckdb/common/enums/metric_type.hpp"
#include "duckdb/common/enum_util.hpp"

namespace duckdb {

"""


def _setup_hpp(out_hpp: Path, f: IndentedFileWriter, metric_index: MetricIndex):
    f.write_header("duckdb/common/enums/metric_type.hpp")
    f.write(HPP_HEADER)

    f.write("enum class MetricGroup : uint8_t {\n")

    groups = metric_index.group_names + ["INVALID"]
    for g in groups:
        f.write_indented(1, f"{g.upper()},")
    f.write("};\n\n")

    f.write("enum class MetricType : uint8_t {\n")
    for g in metric_index.group_names:
        if g == "all" or g == "default":
            continue

        f.write_indented(1, f"// {_to_pascal_case(g)} metrics")
        for m in metric_index.metrics_per_group(g):
            val = metric_index.metric_value(m)
            if val is None:
                raise ValueError(f"Metric '{m}' in group '{g}' has no explicit value assigned")
            f.write_indented(1, f"{m.upper()} = {val},")

    f.write("};\n")

    f.write(HPP_TYPEDEFS)
    f.write('class MetricsUtils {\n')
    f.write('public:\n')
    opt_metrics = metric_index.metrics_per_group("optimizer")
    if opt_metrics:
        f.write_indented(
            1,
            f"static constexpr uint8_t START_OPTIMIZER = static_cast<uint8_t>(MetricType::{opt_metrics[0]});",
        )
        f.write_indented(
            1,
            f"static constexpr uint8_t END_OPTIMIZER = static_cast<uint8_t>(MetricType::{opt_metrics[-1]});\n",
        )
    f.write('public:\n')


def _generate_standard_functions(
    group: str, hpp_f: IndentedFileWriter, cpp_f: IndentedFileWriter, metric_index: MetricIndex
):
    formatted = _to_pascal_case(group)
    get_fn = f"Get{formatted}Metrics"

    hpp_f.write('\n')
    hpp_f.write_indented(1, f"// {formatted} metrics")
    hpp_f.write_indented(1, f"static profiler_settings_t {get_fn}();")

    metrics = metric_index.metrics_per_group(group) if group != "root_scope" else metric_index.root_scope_metrics()

    cpp_f.write(f"profiler_settings_t MetricsUtils::{get_fn}() {{\n")

    if group == "optimizer":
        cpp_f.write_indented(1, "profiler_settings_t result;")
        cpp_f.write_indented(1, f"for (auto metric = START_OPTIMIZER; metric <= END_OPTIMIZER; metric++) {{")
        cpp_f.write_indented(2, f"result.insert(static_cast<MetricType>(metric));")
        cpp_f.write_indented(1, "}")
        cpp_f.write_indented(1, "return result;")
    else:
        cpp_f.write_indented(1, "return {")
        for m in metrics:
            cpp_f.write_indented(2, f"MetricType::{m},")
        cpp_f.write_indented(1, "};")
    cpp_f.write('}\n\n')

    if group == "all":
        _generate_get_metric_by_group_function(hpp_f, cpp_f, metric_index)
        return

    check_fn = f"Is{formatted}Metric"
    hpp_f.write_indented(1, f"static bool {check_fn}(MetricType type);")

    cpp_f.write(f"bool MetricsUtils::{check_fn}(MetricType type) {{\n")
    if group == "optimizer":
        cpp_f.write_indented(
            1,
            f"return static_cast<uint8_t>(type) >= START_OPTIMIZER && static_cast<uint8_t>(type) <= END_OPTIMIZER;",
        )
    else:
        cpp_f.write_indented(1, "switch(type) {")
        for m in metrics:
            cpp_f.write_indented(1, f"case MetricType::{m}:")
        cpp_f.write_indented(2, "return true;")
        cpp_f.write_indented(1, "default:")
        cpp_f.write_indented(2, "return false;")
        cpp_f.write_indented(1, "}")
    cpp_f.write("}\n\n")


def _generate_custom_optimizer_functions(
    optimizers: List[Tuple[str, int]], hpp_f: IndentedFileWriter, cpp_f: IndentedFileWriter
):
    by_type = "GetOptimizerMetricByType(OptimizerType type)"
    by_metric = "GetOptimizerTypeByMetric(MetricType type)"

    hpp_f.write_indented(1, f"static MetricType {by_type};")
    hpp_f.write_indented(1, f"static OptimizerType {by_metric};")

    first_optimizer = optimizers[0][0]

    cpp_f.write(
        f"""
MetricType MetricsUtils::GetOptimizerMetricByType(OptimizerType type) {{
	if (type == OptimizerType::INVALID) {{
		throw InternalException("Invalid OptimizerType: INVALID");
	}}

	const auto base_opt = static_cast<uint8_t>(OptimizerType::{first_optimizer});
	const auto idx = static_cast<uint8_t>(type) - base_opt;

	const auto metric_u8 = static_cast<uint8_t>(START_OPTIMIZER + idx);
	if (metric_u8 < START_OPTIMIZER || metric_u8 > END_OPTIMIZER) {{
		throw InternalException("OptimizerType out of MetricType optimizer range");
	}}
	return static_cast<MetricType>(metric_u8);
}}

OptimizerType MetricsUtils::GetOptimizerTypeByMetric(MetricType type) {{
	const auto metric_u8 = static_cast<uint8_t>(type);
	if (!IsOptimizerMetric(type)) {{
		throw InternalException("MetricType is not an optimizer metric");
	}}

	const auto idx = static_cast<uint8_t>(metric_u8 - START_OPTIMIZER);
	const auto result = static_cast<uint8_t>(OptimizerType::{first_optimizer}) + idx;
	return static_cast<OptimizerType>(result);
}}

"""
    )


def _generate_get_metric_by_group_function(
    hpp_f: IndentedFileWriter, cpp_f: IndentedFileWriter, metric_index: MetricIndex
):
    fn = "GetMetricsByGroupType(MetricGroup type)"
    hpp_f.write_indented(1, f"static profiler_settings_t {fn};")

    cpp_f.write(f"profiler_settings_t MetricsUtils::{fn} {{\n")
    cpp_f.write_indented(1, "switch(type) {")
    for group in metric_index.group_names:
        formatted = group.upper()
        cpp_f.write_indented(1, f"case MetricGroup::{formatted}:")
        cpp_f.write_indented(2, "return Get" + _to_pascal_case(group) + "Metrics();")
    cpp_f.write_indented(1, "default:")
    cpp_f.write_indented(2, 'throw InternalException("The MetricGroup passed is invalid");')
    cpp_f.write_indented(1, "}")
    cpp_f.write('}\n\n')


def generate_metric_type_files(
    out_hpp: Path,
    out_cpp: Path,
    metric_index: MetricIndex,
    optimizers: List[Tuple[str, int]],
) -> None:
    with IndentedFileWriter(out_hpp) as hpp_f, IndentedFileWriter(out_cpp) as cpp_f:
        _setup_hpp(out_hpp, hpp_f, metric_index)
        cpp_f.write(write_warning())
        cpp_f.write(CPP_HEADER)

        for group in metric_index.metrics_by_group:
            _generate_standard_functions(group, hpp_f, cpp_f, metric_index)
            if group == "optimizer":
                _generate_custom_optimizer_functions(optimizers, hpp_f, cpp_f)

        _generate_standard_functions("root_scope", hpp_f, cpp_f, metric_index)

        hpp_f.write("};\n")
        hpp_f.write("} // namespace duckdb\n")
        cpp_f.write("}\n")



# fmt: off

import pytest
import subprocess
import sys
from typing import List
from conftest import ShellTest
import os

def test_readable_numbers(shell):
    test = (
        ShellTest(shell)
        .statement(".large_number_rendering footer")
        .statement("select 59986052 as count, 123456789 as count2, 9999999999 count3, -9999999999 count4;")
    )
    result = test.run()
    result.check_stdout("(59.99 million)")
    result.check_stdout("(123.46 million)")
    result.check_stdout("(10.00 billion)")
    result.check_stdout("(-10.00 billion)")

@pytest.mark.parametrize('test_rounding', [False, True])
def test_readable_numbers_exhaustive(shell, test_rounding):
    query = "select "
    for i in range(1, 20):
        if i > 1:
            query += ", "
        if test_rounding:
            query += '9' * i
        else:
            query += '1' + ('0' * i)
    test = (
        ShellTest(shell)
        .statement(".large_number_rendering all")
        .statement(".maxwidth 99999")
        .statement(query)
    )
    result = test.run()
    for unit in ['million', 'billion', 'trillion', 'quadrillion', 'quintillion']:
        for number in ['1.00', '10.00', '100.00']:
            if unit == 'quintillion' and number in ['10.00', '100.00']:
                continue
            result.check_stdout(number + " " + unit)

def test_readable_numbers_rounding(shell):
    test = (
        ShellTest(shell)
        .statement(".large_number_rendering footer")
        .statement(".maxwidth 99999")
        .statement("select 1005000, 1004999, -1005000, -1004999;")
    )
    result = test.run()
    result.check_stdout("(1.01 million)")
    result.check_stdout("(1.00 million)")
    result.check_stdout("(-1.01 million)")
    result.check_stdout("(-1.00 million)")

def test_readable_rounding_edge_case(shell):
    test = (
        ShellTest(shell)
        .statement(".large_number_rendering all")
        .statement(".maxwidth 99999")
        .statement("select 994999, 995000")
    )
    result = test.run()
    result.check_stdout("1.00 million")

def test_readable_numbers_limit(shell):
    test = (
        ShellTest(shell)
        .statement(".maxwidth 99999")
        .statement(".large_number_rendering all")
        .statement("select 18446744073709551616, -18446744073709551616, 9999999999999999999, -9999999999999999999;")
    )
    result = test.run()
    result.check_stdout("10.00 quintillion")
    result.check_stdout("-10.00 quintillion")

def test_decimal_separator(shell):
    test = (
        ShellTest(shell)
        .statement(".decimal_sep ,")
        .statement(".large_number_rendering all")
        .statement("select 59986052, 59986052.5, 999999999.123456789, 1e20, 'nan'::double;")
    )
    result = test.run()
    result.check_stdout("59,99 million")
    result.check_stdout("1,00 billion")

def test_odd_floating_points(shell):
    test = (
        ShellTest(shell)
        .statement("select 1e20, 'nan'::double;")
    )
    result = test.run()
    result.check_stdout("nan")

def test_disable_readable_numbers(shell):
    test = (
        ShellTest(shell)
        .statement(".large_number_rendering off")
        .statement("select 123456789;")
    )
    result = test.run()
    result.check_not_exist('(123.46 million)')

def test_large_number_rendering_all(shell):
    test = (
        ShellTest(shell)
        .statement(".large_number_rendering all")
        .statement("select 123456789 from range(10);")
    )
    result = test.run()
    result.check_stdout('123.46 million')
    result.check_not_exist('(123.46 million)')

def test_readable_numbers_columns(shell):
    test = (
        ShellTest(shell)
        .statement(".columns")
        .statement("select 123456789;")
    )
    result = test.run()
    result.check_not_exist('(123.46 million)')

def test_readable_numbers_row_count(shell):
    test = (
        ShellTest(shell)
        .statement(".large_number_rendering footer")
        .statement("select r from range(1230000) t(r);")
    )
    result = test.run()
    result.check_stdout('1.23 million rows')

def test_readable_numbers_row_count_wide(shell):
    test = (
        ShellTest(shell)
        .statement(".large_number_rendering footer")
        .statement("select r, r, r, r, r, r, r from range(1230000) t(r);")
    )
    result = test.run()
    result.check_stdout('1.23 million rows')

def test_readable_numbers_row_count_wide_single_col(shell):
    test = (
        ShellTest(shell)
        .statement(".large_number_rendering footer")
        .statement("select concat(r, r, r, r, r, r, r) c from range(1230000) t(r);")
    )
    result = test.run()
    result.check_stdout('1.23 million rows')



# fmt: off

from conftest import ShellTest

def test_get_env(shell):
    test = (
        ShellTest(shell)
            .statement('.null NULL')
            .statement("SET default_null_order=getenv('DEFAULT_NULL_ORDER');")
            .statement("SELECT * FROM (VALUES (42), (NULL)) ORDER BY 1 LIMIT 1;")
    )
    test.environment['DEFAULT_NULL_ORDER'] = 'NULLS_FIRST'
    result = test.run()
    result.check_stdout('NULL')

    test.environment['DEFAULT_NULL_ORDER'] = 'NULLS_LAST'
    result = test.run()
    result.check_stdout('42')

def test_get_env_permissions(shell):
    test = (
        ShellTest(shell)
            .statement('SET enable_external_access=false')
            .statement("SELECT getenv('DEFAULT_NULL_ORDER');")
    )
    test.environment['DEFAULT_NULL_ORDER'] = 'NULLS_FIRST'
    result = test.run()
    result.check_stderr('disabled through configuration')

# fmt: on



#!/usr/bin/env python3
"""
Exports the BWC test cache from runtime directories.

For each version, creates a tar.gz archive containing only the files needed
for cached test runs:
  - *.sql (generated query files)
  - *.plan.bin (serialized execution plans)
  - *.<version>.result.bin (old version results)
  - fixtures/ directories (test fixture databases)

Usage:
  python3 test/bwc/export_cache.py [--version v1.4.4] [--output-dir /path/to/output]

By default, exports all versions found in the runtime directory.
"""

import argparse
import os
import tarfile
import sys
from os.path import abspath, dirname

parser = argparse.ArgumentParser(description='Export BWC test cache from runtime directories')
parser.add_argument('--version', dest='version', help='Export cache for a specific version only')
parser.add_argument(
    '--output-dir',
    dest='output_dir',
    help='Output directory for cache archives (default: current directory)',
    default='.',
)
parser.add_argument('--runtime-dir', dest='runtime_dir', help='Path to the BWC runtime directory', default=None)

args = parser.parse_args()

duckdb_root_dir = dirname(dirname(dirname(abspath(__file__))))


def get_runtime_dir():
    if args.runtime_dir:
        return args.runtime_dir
    return os.path.join(duckdb_root_dir, 'duckdb_unittest_tempdir', 'bwc', 'runtime')


def should_include(file_path, version):
    """Check if a file should be included in the cache archive."""
    basename = os.path.basename(file_path)
    # Include SQL query files
    if basename.endswith('.sql'):
        return True
    # Include serialized plans
    if basename.endswith('.plan.bin'):
        return True
    # Include old version results (e.g., test.v1.4.4.result.bin)
    if basename.endswith(f'.{version}.result.bin'):
        return True
    return False


def should_include_dir(dir_name):
    """Check if a directory should be included."""
    return dir_name == 'fixtures'


def export_version(version, runtime_dir, output_dir):
    version_dir = os.path.join(runtime_dir, version)
    if not os.path.isdir(version_dir):
        print(f"Warning: runtime directory not found for {version}: {version_dir}")
        return None

    archive_name = f"runtime_{version}.tar.gz"
    archive_path = os.path.join(output_dir, archive_name)

    file_count = 0
    with tarfile.open(archive_path, 'w:gz') as tar:
        for root, dirs, files in os.walk(version_dir):
            # Skip output directories and symlinks
            dirs[:] = [d for d in dirs if not d.startswith('output') and not os.path.islink(os.path.join(root, d))]

            rel_root = os.path.relpath(root, runtime_dir)

            # Include fixture directories entirely
            fixture_dirs = [d for d in dirs if should_include_dir(d)]
            for d in fixture_dirs:
                fixture_path = os.path.join(root, d)
                arcname = os.path.join(rel_root, d)
                tar.add(fixture_path, arcname=arcname)
                file_count += sum(len(f) for _, _, f in os.walk(fixture_path))
                # Don't recurse into fixtures again
                dirs.remove(d)

            for file in files:
                file_path = os.path.join(root, file)
                if should_include(file_path, version):
                    arcname = os.path.join(rel_root, file)
                    tar.add(file_path, arcname=arcname)
                    file_count += 1

    size_mb = os.path.getsize(archive_path) / (1024 * 1024)
    print(f"Exported {version}: {file_count} files -> {archive_path} ({size_mb:.1f} MB)")
    return archive_path


if __name__ == '__main__':
    runtime_dir = get_runtime_dir()
    if not os.path.isdir(runtime_dir):
        print(f"Error: runtime directory not found: {runtime_dir}")
        sys.exit(1)

    os.makedirs(args.output_dir, exist_ok=True)

    if args.version:
        versions = [args.version]
    else:
        versions = sorted(
            [d for d in os.listdir(runtime_dir) if d.startswith('v') and os.path.isdir(os.path.join(runtime_dir, d))]
        )

    if not versions:
        print(f"No version directories found in {runtime_dir}")
        sys.exit(1)

    print(f"Exporting cache for {len(versions)} version(s) from {runtime_dir}")
    exported = []
    for version in versions:
        result = export_version(version, runtime_dir, args.output_dir)
        if result:
            exported.append(result)

    print(f"\nDone: exported {len(exported)}/{len(versions)} archives to {args.output_dir}")



# use flex to generate the scanner file for the parser
# the following version of bison is used:
# flex 2.5.35 Apple(flex-32)
import os
import subprocess
import re
from sys import platform
import sys
from python_helpers import open_utf8

flex_bin = 'flex'
pg_path = os.path.join('third_party', 'libpg_query')
namespace = 'duckdb_libpgquery'

for arg in sys.argv[1:]:
    if arg.startswith("--flex="):
        flex_bin = arg.replace("--flex=", "")
    elif arg.startswith("--custom_dir_prefix"):
        pg_path = arg.split("=")[1] + pg_path
    elif arg.startswith("--namespace"):
        namespace = arg.split("=")[1]
    else:
        raise Exception("Unrecognized argument: " + arg + ", expected --flex, --custom_dir_prefix, --namespace")

flex_file_path = os.path.join(pg_path, 'scan.l')
target_file = os.path.join(pg_path, 'src_backend_parser_scan.cpp')

proc = subprocess.Popen(
    [flex_bin, '--nounistd', '-o', target_file, flex_file_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE
)
stdout = proc.stdout.read().decode('utf8')
stderr = proc.stderr.read().decode('utf8')
if proc.returncode != None or len(stderr) > 0:
    print("Flex failed")
    print("stdout: ", stdout)
    print("stderr: ", stderr)
    exit(1)

with open_utf8(target_file, 'r') as f:
    text = f.read()

# convert this from 'int' to 'yy_size_t' to avoid triggering a warning
text = text.replace('int yy_buf_size;\n', 'yy_size_t yy_buf_size;\n')

# add the libpg_query namespace
text = text.replace(
    '''
#ifndef FLEXINT_H
#define FLEXINT_H
''',
    '''
#ifndef FLEXINT_H
#define FLEXINT_H
namespace '''
    + namespace
    + ''' {
''',
)
text = text.replace('register ', '')

text = text + "\n} /* " + namespace + " */\n"

text = re.sub('(?:[(]void[)][ ]*)?fprintf', '//', text)
text = re.sub('exit[(]', 'throw std::runtime_error(msg); //', text)
text = re.sub(r'\n\s*if\s*[(]\s*!\s*yyin\s*[)]\s*\n\s*yyin\s*=\s*stdin;\s*\n', '\n', text)
text = re.sub(r'\n\s*if\s*[(]\s*!\s*yyout\s*[)]\s*\n\s*yyout\s*=\s*stdout;\s*\n', '\n', text)

file_null = 'NULL' if platform == 'linux' else '[(]FILE [*][)] 0'

text = re.sub(
    rf'[#]ifdef\s*YY_STDINIT\n\s*yyin = stdin;\n\s*yyout = stdout;\n[#]else\n\s*yyin = {file_null};\n\s*yyout = {file_null};\n[#]endif',
    '    yyin = (FILE *) 0;\n    yyout = (FILE *) 0;',
    text,
)

if 'stdin;' in text:
    print("STDIN not removed!")
    # exit(1)

if 'stdout' in text:
    print("STDOUT not removed!")
    # exit(1)

if 'fprintf(' in text:
    print("PRINTF not removed!")
    # exit(1)

if 'exit(' in text:
    print("EXIT not removed!")
    # exit(1)

with open_utf8(target_file, 'w+') as f:
    f.write(text)



# fmt: off

import pytest
import subprocess
import sys
import os
from typing import List
from conftest import ShellTest

def test_logging(shell):
    test = (
        ShellTest(shell)
        .statement("CALL enable_logging('QueryLog', storage='stdout')")
        .statement('SELECT 1 as a')
    )
    result = test.run()

    newline = "\r\n" if os.name == "nt" else "\n"
    result.check_stdout(f"QueryLog\tINFO\tSELECT 1 as a;{newline}┌───────┐")


def test_logging_custom_delim(shell):
    test = (
        ShellTest(shell)
        .statement("CALL enable_logging('QueryLog', storage='stdout', storage_config={'delim':','})")
        .statement('SELECT 1 as a')
    )
    result = test.run()
    newline = "\r\n" if os.name == "nt" else "\n"
    result.check_stdout(f"QueryLog,INFO,SELECT 1 as a;{newline}┌───────┐")

# By default stdoutlogging has buffer size of 1, but we can increase it if we want. We use `only_flush_on_full_buffer` to ensure we can test this
def test_logging_buffering(shell):
    test = (
        ShellTest(shell)
        .statement("CALL enable_logging('QueryLog', storage='stdout', storage_buffer_size=1000, storage_config={'only_flush_on_full_buffer': true})")
        .statement('SELECT 1 as a')
        .statement('SELECT 2 as b')
    )
    result = test.run()
    result.check_not_exist("QueryLog")

# fmt: on
